import gzip
import json
import os
import unittest

from torch_tb_profiler.profiler.data import (DistributedRunProfileData,
                                             RunProfileData)
from torch_tb_profiler.profiler.loader import RunLoader
from torch_tb_profiler.profiler.overall_parser import ProfileRole
from torch_tb_profiler.profiler.gpu_metrics_parser import GPUMetricsParser
from torch_tb_profiler.run import RunProfile

SCHEMA_VERSION = 1
WORKER_NAME = 'worker0'


def parse_json_trace(json_content, worker_name=WORKER_NAME) -> RunProfileData:
    trace_json = json.loads(json_content)
    trace_json = {'schemaVersion': 1, 'traceEvents': trace_json}
    return RunProfileData.from_json(worker_name, 0, trace_json)


'''
All the events in json string are only simulation, not actual generated events.
We removed the data fields that not used by current version of our profiler,
for easy to check correctness and shorter in length.
We even renamed the data values such as kernel name or 'ts', to simplify the string.
'''


class TestProfiler(unittest.TestCase):
    #  A test case including all 7 event categories.
    def test_all_categories(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "enumerate(DataLoader)#_SingleProcessDataLoaderIter.__next__", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 180,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 200, "dur": 60,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::nll_loss_backward", "pid": 13721, "tid": "456",
                "ts": 340, "dur": 70,
                "args": {"Input Dims": [[], [32, 1000], [32], [], [], [], []], "External id": 4}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#1", "pid": 13721, "tid": "123",
                "ts": 50, "dur": 400,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Memcpy",
                "name": "Memcpy HtoD (Pageable -> Device)", "pid": 0, "tid": "stream 7",
                "ts": 405, "dur": 10,
                "args": {"stream": 7, "correlation": 334, "external id": 4}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaMemcpyAsync", "pid": 13721, "tid": "456",
                "ts": 360, "dur": 20,
                "args": {"correlation": 334, "external id": 4}
            },
            {
                "ph": "X", "cat": "Memset",
                "name": "Memset (Device)", "pid": 0, "tid": "stream 7",
                "ts": 420, "dur": 5,
                "args": {"stream": 7, "correlation": 40344, "external id": 4}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaMemsetAsync", "pid": 13721, "tid": "456",
                "ts": 390, "dur": 10,
                "args": {"correlation": 40344, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 430, "dur": 15,
                "args": {"correlation": 40348, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 405, "dur": 5,
                "args": {"correlation": 40348, "external id": 4}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        self.assertTrue(profile.has_runtime)
        self.assertTrue(profile.has_kernel)
        self.assertTrue(profile.has_memcpy_or_memset)
        step = profile.steps_costs[0]
        self.assertEqual(step.costs[ProfileRole.Kernel], 15)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 10)
        self.assertEqual(step.costs[ProfileRole.Memset], 5)
        self.assertEqual(step.costs[ProfileRole.Runtime], 30)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 180)
        self.assertEqual(step.costs[ProfileRole.CpuOp], 35)
        self.assertEqual(step.costs[ProfileRole.Other], 125)

        self.assertEqual(len(profile.op_list_groupby_name), 2)
        self.assertEqual(len(profile.op_list_groupby_name_input), 2)

        def test_op_list(op_list):
            op_count = 0
            for op_agg in op_list:
                if op_agg.name == 'aten::to':
                    op_count += 1
                    self.assertEqual(op_agg.input_shape,
                                     '[[2, 8, 5], [], [], [], [], [], [], []]')
                    self.assertEqual(op_agg.calls, 1)
                    self.assertEqual(op_agg.host_duration, 60)
                    self.assertEqual(op_agg.device_duration, 0)
                    self.assertEqual(op_agg.self_host_duration, 60)
                    self.assertEqual(op_agg.self_device_duration, 0)
                if op_agg.name == 'aten::nll_loss_backward':
                    op_count += 1
                    self.assertEqual(op_agg.input_shape,
                                     '[[], [32, 1000], [32], [], [], [], []]')
                    self.assertEqual(op_agg.calls, 1)
                    self.assertEqual(op_agg.host_duration, 70)
                    self.assertEqual(op_agg.device_duration, 30)
                    self.assertEqual(
                        op_agg.self_host_duration, 70 - 20 - 10 - 5)
                    self.assertEqual(op_agg.self_device_duration, 30)
            self.assertEqual(op_count, 2)

        test_op_list(profile.op_list_groupby_name)
        test_op_list(profile.op_list_groupby_name_input)

        self.assertEqual(len(profile.kernel_list_groupby_name_op), 1)
        self.assertEqual(profile.kernel_stat.shape[0], 1)
        self.assertEqual(profile.kernel_list_groupby_name_op[0].name,
                         'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>')
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].op_name, 'aten::nll_loss_backward')
        self.assertEqual(profile.kernel_list_groupby_name_op[0].calls, 1)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].total_duration, 15)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].min_duration, 15)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].max_duration, 15)
        self.assertEqual(profile.kernel_stat.iloc[0]['count'], 1)
        self.assertEqual(profile.kernel_stat.iloc[0]['sum'], 15)
        self.assertEqual(profile.kernel_stat.iloc[0]['mean'], 15)
        self.assertEqual(profile.kernel_stat.iloc[0]['min'], 15)
        self.assertEqual(profile.kernel_stat.iloc[0]['max'], 15)

    # Test using external_id to build relationship between Operator and Runtime.
    # Use external_id to build correlation with its father OperatorNode or ProfilerStepNode.
    # Because in the case when RuntimeNode has duration 0 and starts at same time as a OperatorNode,
    # just use interval containing relationship can't tell it is child or brother of the OperatorNode.
    def test_external_id(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "aten::mat_mul", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 70,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 130, "dur": 5,
                "args": {"correlation": 334, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 0,
                "args": {"correlation": 334, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 130, "dur": 6,
                "args": {"correlation": 335, "external id": 2, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 0,
                "args": {"correlation": 335, "external id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 130, "dur": 7,
                "args": {"correlation": 336, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 190, "dur": 0,
                "args": {"correlation": 336, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 130, "dur": 8,
                "args": {"correlation": 337, "external id": 2, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 190, "dur": 0,
                "args": {"correlation": 337, "external id": 2}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        op_count = 0
        for op_agg in profile.op_list_groupby_name:
            if op_agg.name == 'aten::mat_mul':
                op_count += 1
                self.assertEqual(op_agg.device_duration, 5 + 6 + 7 + 8)
                self.assertEqual(op_agg.self_device_duration, 6 + 8)
            if op_agg.name == 'aten::mm':
                op_count += 1
                self.assertEqual(op_agg.device_duration, 5 + 7)
                self.assertEqual(op_agg.self_device_duration, 5 + 7)
        self.assertEqual(op_count, 2)

    # Test operator's father-child relationship when they have same start time or end time.
    def test_operator_relation(self):
        # 2 events with same start time.
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "aten::mat_mul", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 70,
                "args": {"Input Dims": [], "External id": 4}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        op_count = 0
        for op_agg in profile.op_list_groupby_name:
            if op_agg.name == 'aten::mat_mul':
                op_count += 1
                self.assertEqual(op_agg.self_host_duration, 100 - 70)
            if op_agg.name == 'aten::mm':
                op_count += 1
                self.assertEqual(op_agg.self_host_duration, 70)
        self.assertEqual(op_count, 2)

        # 2 events with same end time.
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "aten::mat_mul", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 130, "dur": 70,
                "args": {"Input Dims": [], "External id": 4}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        op_count = 0
        for op_agg in profile.op_list_groupby_name:
            if op_agg.name == 'aten::mat_mul':
                op_count += 1
                self.assertEqual(op_agg.self_host_duration, 100 - 70)
            if op_agg.name == 'aten::mm':
                op_count += 1
                self.assertEqual(op_agg.self_host_duration, 70)
        self.assertEqual(op_count, 2)

    # Test multiple father-child operators with same name.
    # In this case, all the operators except the top operator should be removed,
    # and all runtime/kernels belong to the children operators should be attached to the only kept one.
    # This behavior is to keep consistent with _remove_dup_nodes in torch/autograd/profiler.py.
    def test_remove_dup_nodes(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 110, "dur": 80,
                "args": {"Input Dims": [], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 60,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 130, "dur": 20,
                "args": {"correlation": 335, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void gemmSN_TN_kernel_64addr", "pid": 0, "tid": "stream 7",
                "ts": 220, "dur": 8,
                "args": {"correlation": 335, "external id": 4, "device": 0}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        self.assertEqual(len(profile.op_list_groupby_name), 1)
        self.assertEqual(
            profile.op_list_groupby_name[0].self_device_duration, 8)

    # Test Runtime with 'external id' 0.
    # This kind of Runtime should not be attached to any operator,
    # and should be included in accumulating device time.
    def test_top_level_runtime(self):
        # This operator is different thread with the runtime.
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 130, "dur": 20,
                "args": {"correlation": 335, "external id": 0}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void gemmSN_TN_kernel_64addr", "pid": 0, "tid": "stream 7",
                "ts": 220, "dur": 8,
                "args": {"correlation": 335, "external id": 0, "device": 0}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        self.assertEqual(profile.op_list_groupby_name[0].device_duration, 0)
        self.assertEqual(
            profile.op_list_groupby_name[0].self_device_duration, 0)
        self.assertEqual(profile.kernel_stat.iloc[0]['count'], 1)

    # Test Runtime directly called in ProfilerStep, not inside any operator.
    def test_runtime_called_by_profilerstep(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#1", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 300,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 130, "dur": 20,
                "args": {"correlation": 335, "external id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void gemmSN_TN_kernel_64addr", "pid": 0, "tid": "stream 7",
                "ts": 220, "dur": 8,
                "args": {"correlation": 335, "external id": 2, "device": 0}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        step = profile.steps_costs[0]
        self.assertEqual(step.costs[ProfileRole.Kernel], 8)
        self.assertEqual(step.costs[ProfileRole.Runtime], 20)
        self.assertEqual(step.costs[ProfileRole.CpuOp], 0)
        self.assertEqual(step.costs[ProfileRole.Other], 300 - 8 - 20)
        # ProfilerStep is not regarded as an operator.
        self.assertEqual(len(profile.op_list_groupby_name), 0)
        self.assertEqual(len(profile.op_list_groupby_name_input), 0)
        self.assertEqual(profile.kernel_stat.iloc[0]['count'], 1)
        self.assertEqual(len(profile.kernel_list_groupby_name_op), 1)

    # Test one Runtime lauch more than one Kernels.
    # Sometimes such as running Bert using DataParallel mode(1 process, 2GPUs),
    # one runtime such as cudaLaunchCooperativeKernelMultiDevice could trigger more than one kernel,
    # each Kernel runs at a seperate GPU card.
    def test_runtime_launch_multipe_kernels(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "Broadcast", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 300,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchCooperativeKernelMultiDevice", "pid": 13721, "tid": "456",
                "ts": 130, "dur": 20,
                "args": {"correlation": 335, "external id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "ncclBroadcastRingLLKernel_copy_i8(ncclColl)", "pid": 0, "tid": "stream 13",
                "ts": 160, "dur": 120318,
                "args": {"device": 0, "context": 1, "stream": 13,
                        "correlation": 335, "external id": 2, "device": 0}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "ncclBroadcastRingLLKernel_copy_i8(ncclColl)", "pid": 0, "tid": "stream 22",
                "ts": 170, "dur": 132800,
                "args": {"device": 0, "context": 2, "stream": 22,
                        "correlation": 335, "external id": 2}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        self.assertEqual(
            profile.op_list_groupby_name[0].device_duration, 120318 + 132800)
        self.assertEqual(profile.kernel_stat.iloc[0]['count'], 2)
        self.assertEqual(len(profile.kernel_list_groupby_name_op), 1)

    # Test when there is no ProfilerStep#.
    def test_no_profilerstep(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 60,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::nll_loss_backward", "pid": 13721, "tid": "456",
                "ts": 300, "dur": 70,
                "args": {"Input Dims": [[], [32, 1000], [32], [], [], [], []], "External id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 320, "dur": 100,
                "args": {"correlation": 40348, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 310, "dur": 20,
                "args": {"correlation": 40348, "external id": 4}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        self.assertTrue(profile.has_runtime)
        self.assertTrue(profile.has_kernel)
        self.assertTrue(not profile.has_memcpy_or_memset)
        self.assertEqual(len(profile.steps_costs), 1)
        step = profile.steps_costs[0]

        self.assertEqual(step.costs[ProfileRole.Kernel], 100)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 0)
        self.assertEqual(step.costs[ProfileRole.Memset], 0)
        self.assertEqual(step.costs[ProfileRole.Runtime], 320 - 310)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 0)
        self.assertEqual(step.costs[ProfileRole.CpuOp], 60 + (310 - 300))
        # If no ProfilerStep, all events will be regarded as a step.
        self.assertEqual(step.costs[ProfileRole.Other], 300 - (100 + 60))
        self.assertEqual(step.costs[ProfileRole.Total], (320 + 100) - 100)
        self.assertEqual(len(profile.op_list_groupby_name), 2)
        self.assertEqual(len(profile.op_list_groupby_name_input), 2)
        self.assertEqual(profile.kernel_stat.iloc[0]['count'], 1)
        self.assertEqual(len(profile.kernel_list_groupby_name_op), 1)

        def test_op_list(op_list):
            op_count = 0
            for op_agg in op_list:
                if op_agg.name == 'aten::to':
                    op_count += 1
                    self.assertEqual(op_agg.input_shape,
                                     '[[2, 8, 5], [], [], [], [], [], [], []]')
                    self.assertEqual(op_agg.calls, 1)
                    self.assertEqual(op_agg.host_duration, 60)
                    self.assertEqual(op_agg.device_duration, 0)
                    self.assertEqual(op_agg.self_host_duration, 60)
                    self.assertEqual(op_agg.self_device_duration, 0)
                if op_agg.name == 'aten::nll_loss_backward':
                    op_count += 1
                    self.assertEqual(op_agg.input_shape,
                                     '[[], [32, 1000], [32], [], [], [], []]')
                    self.assertEqual(op_agg.calls, 1)
                    self.assertEqual(op_agg.host_duration, 70)
                    self.assertEqual(op_agg.device_duration, 100)
                    self.assertEqual(op_agg.self_host_duration, 70 - 20)
                    self.assertEqual(op_agg.self_device_duration, 100)
            self.assertEqual(op_count, 2)

        test_op_list(profile.op_list_groupby_name)
        test_op_list(profile.op_list_groupby_name_input)

        self.assertEqual(profile.kernel_list_groupby_name_op[0].name,
                         'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>')
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].op_name, 'aten::nll_loss_backward')
        self.assertEqual(profile.kernel_list_groupby_name_op[0].calls, 1)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].total_duration, 100)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].min_duration, 100)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].max_duration, 100)
        self.assertEqual(profile.kernel_stat.iloc[0]['count'], 1)
        self.assertEqual(profile.kernel_stat.iloc[0]['sum'], 100)
        self.assertEqual(profile.kernel_stat.iloc[0]['mean'], 100)
        self.assertEqual(profile.kernel_stat.iloc[0]['min'], 100)
        self.assertEqual(profile.kernel_stat.iloc[0]['max'], 100)

    # 2 steps without overlap with each other.
    def test_multiple_profilersteps_no_overlap(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#1", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 200,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 200, "dur": 60,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#2", "pid": 13721, "tid": "123",
                "ts": 350, "dur": 150,
                "args": {"Input Dims": [], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "123",
                "ts": 360, "dur": 50,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Memcpy",
                "name": "Memcpy HtoD (Pageable -> Device)", "pid": 0, "tid": "stream 7",
                "ts": 280, "dur": 40,
                "args": {"stream": 7, "correlation": 334, "external id": 2}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaMemcpyAsync", "pid": 13721, "tid": "123",
                "ts": 250, "dur": 5,
                "args": {"correlation": 334, "external id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 410, "dur": 200,
                "args": {"correlation": 40348, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "123",
                "ts": 400, "dur": 5,
                "args": {"correlation": 40348, "external id": 4}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        self.assertTrue(profile.has_runtime)
        self.assertTrue(profile.has_kernel)
        self.assertTrue(profile.has_memcpy_or_memset)
        self.assertEqual(len(profile.steps_costs), 2)
        step = profile.steps_costs[0]
        self.assertEqual(step.costs[ProfileRole.Kernel], 0)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 40)
        self.assertEqual(step.costs[ProfileRole.Memset], 0)
        self.assertEqual(step.costs[ProfileRole.Runtime], 5)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 0)
        self.assertEqual(step.costs[ProfileRole.CpuOp], 60 - 5)
        self.assertEqual(step.costs[ProfileRole.Other], 200 - 60 - 20)
        # Device side takes effect.
        self.assertEqual(step.costs[ProfileRole.Total], 320 - 100)
        step = profile.steps_costs[1]
        self.assertEqual(step.costs[ProfileRole.Kernel], 200)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 0)
        self.assertEqual(step.costs[ProfileRole.Memset], 0)
        self.assertEqual(step.costs[ProfileRole.Runtime], 5)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 0)
        self.assertEqual(step.costs[ProfileRole.CpuOp], 50 - 5)
        self.assertEqual(step.costs[ProfileRole.Other], 360 - 350)
        # Device side takes effect.
        self.assertEqual(step.costs[ProfileRole.Total], 610 - 350)
        self.assertEqual(
            profile.avg_costs.costs[ProfileRole.Total], ((320 - 100) + (610 - 350)) / 2)

        self.assertEqual(len(profile.op_list_groupby_name), 2)
        self.assertEqual(len(profile.op_list_groupby_name_input), 2)

        def test_op_list(op_list):
            op_count = 0
            for op_agg in op_list:
                if op_agg.name == 'aten::to':
                    op_count += 1
                    self.assertEqual(op_agg.input_shape,
                                     '[[2, 8, 5], [], [], [], [], [], [], []]')
                    self.assertEqual(op_agg.calls, 1)
                    self.assertEqual(op_agg.host_duration, 60)
                    self.assertEqual(op_agg.device_duration, 40)
                    self.assertEqual(op_agg.self_host_duration, 60 - 5)
                    self.assertEqual(op_agg.self_device_duration, 40)
                if op_agg.name == 'aten::mm':
                    op_count += 1
                    self.assertEqual(op_agg.input_shape, '[]')
                    self.assertEqual(op_agg.calls, 1)
                    self.assertEqual(op_agg.host_duration, 50)
                    self.assertEqual(op_agg.device_duration, 200)
                    self.assertEqual(op_agg.self_host_duration, 50 - 5)
                    self.assertEqual(op_agg.self_device_duration, 200)
            self.assertEqual(op_count, 2)

        test_op_list(profile.op_list_groupby_name)
        test_op_list(profile.op_list_groupby_name_input)

        self.assertEqual(len(profile.kernel_list_groupby_name_op), 1)
        self.assertEqual(profile.kernel_stat.shape[0], 1)
        self.assertEqual(profile.kernel_list_groupby_name_op[0].name,
                         'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>')
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].op_name, 'aten::mm')
        self.assertEqual(profile.kernel_list_groupby_name_op[0].calls, 1)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].total_duration, 200)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].min_duration, 200)
        self.assertEqual(
            profile.kernel_list_groupby_name_op[0].max_duration, 200)
        self.assertEqual(profile.kernel_stat.iloc[0]['count'], 1)
        self.assertEqual(profile.kernel_stat.iloc[0]['sum'], 200)
        self.assertEqual(profile.kernel_stat.iloc[0]['mean'], 200)
        self.assertEqual(profile.kernel_stat.iloc[0]['min'], 200)
        self.assertEqual(profile.kernel_stat.iloc[0]['max'], 200)

    # Test self time and total time on operator with nested operator.
    def test_self_time(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "aten::mat_mul", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 40,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 155, "dur": 20,
                "args": {"correlation": 334, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 150, "dur": 10,
                "args": {"correlation": 334, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 210, "dur": 16,
                "args": {"correlation": 335, "external id": 2, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 170, "dur": 25,
                "args": {"correlation": 335, "external id": 2}
            }]
        """
        profile = parse_json_trace(json_content)

        op_count = 0
        for op_agg in profile.op_list_groupby_name:
            if op_agg.name == 'aten::mat_mul':
                op_count += 1
                self.assertEqual(op_agg.host_duration, 100)
                self.assertEqual(op_agg.device_duration, 20 + 16)
                self.assertEqual(op_agg.self_host_duration, 100 - 40 - 25)
                self.assertEqual(op_agg.self_device_duration, 16)
            if op_agg.name == 'aten::mm':
                op_count += 1
                self.assertEqual(op_agg.host_duration, 40)
                self.assertEqual(op_agg.device_duration, 20)
                self.assertEqual(op_agg.self_host_duration, 30)
                self.assertEqual(op_agg.self_device_duration, 20)
        self.assertEqual(op_count, 2)

    # 2 steps with overlap with each other.
    def test_multiple_profilersteps_with_overlap(self):
        # The kernel with 'correlation' as 123 is launched by previous step,
        # its end time is bigger than 'ProfilerStep#1''s start time,
        # so it is regarded as beginning of 'ProfilerStep#1'.
        # The memcpy with 'correlation' as 334 is launched by 'ProfilerStep#1',
        # its end time is bigger than 'ProfilerStep#2''s start time,
        # so it is regarded as beginning of 'ProfilerStep#2'.
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#1", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 200,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 200, "dur": 60,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#2", "pid": 13721, "tid": "123",
                "ts": 350, "dur": 150,
                "args": {"Input Dims": [], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "123",
                "ts": 360, "dur": 50,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 150, "dur": 90,
                "args": {"correlation": 123, "external id": 0, "device": 0}
            },
            {
                "ph": "X", "cat": "Memcpy",
                "name": "Memcpy HtoD (Pageable -> Device)", "pid": 0, "tid": "stream 7",
                "ts": 280, "dur": 100,
                "args": {"stream": 7, "correlation": 334, "external id": 2}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaMemcpyAsync", "pid": 13721, "tid": "123",
                "ts": 250, "dur": 5,
                "args": {"correlation": 334, "external id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 410, "dur": 200,
                "args": {"correlation": 40348, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "123",
                "ts": 400, "dur": 5,
                "args": {"correlation": 40348, "external id": 4}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        self.assertTrue(profile.has_runtime)
        self.assertTrue(profile.has_kernel)
        self.assertTrue(profile.has_memcpy_or_memset)
        self.assertEqual(len(profile.steps_costs), 2)
        step = profile.steps_costs[0]
        self.assertEqual(step.costs[ProfileRole.Kernel], 0)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 100)
        self.assertEqual(step.costs[ProfileRole.Memset], 0)
        self.assertEqual(step.costs[ProfileRole.Runtime], 5)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 0)
        self.assertEqual(step.costs[ProfileRole.CpuOp],
                         (200 + 60) - (150 + 90) - 5)
        self.assertEqual(step.costs[ProfileRole.Other], 280 - (200 + 60))
        # Device side takes effect.
        self.assertEqual(step.costs[ProfileRole.Total],
                         (280 + 100) - (150 + 90))
        step = profile.steps_costs[1]
        self.assertEqual(step.costs[ProfileRole.Kernel], 200)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 0)
        self.assertEqual(step.costs[ProfileRole.Memset], 0)
        self.assertEqual(step.costs[ProfileRole.Runtime], 5)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 0)
        self.assertEqual(step.costs[ProfileRole.CpuOp],
                         (280 + 100) - 360 + (410 - 405))
        self.assertEqual(step.costs[ProfileRole.Other], 0)
        # Device side takes effect.
        self.assertEqual(step.costs[ProfileRole.Total], 610 - (280 + 100))

    # Test whether step time is calculated correctly when the last 2 steps have no kernels launched.
    def test_last_steps_no_kernel(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#1", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 200,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 120, "dur": 10,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#2", "pid": 13721, "tid": "123",
                "ts": 300, "dur": 100,
                "args": {"Input Dims": [], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#3", "pid": 13721, "tid": "123",
                "ts": 400, "dur": 50,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 90, "dur": 20,
                "args": {"correlation": 123, "external id": 0, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaMemcpyAsync", "pid": 13721, "tid": "123",
                "ts": 125, "dur": 5,
                "args": {"correlation": 334, "external id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 150, "dur": 180,
                "args": {"correlation": 334, "external id": 2, "device": 0}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        # The last 2 steps without kernels are removed from overall view.
        self.assertEqual(len(profile.steps_costs), 1)
        step = profile.steps_costs[0]
        self.assertEqual(
            step.costs[ProfileRole.Total], (150 + 180) - (90 + 20))

    def test_pure_cpu(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#1", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 200,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 120, "dur": 10,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#2", "pid": 13721, "tid": "123",
                "ts": 300, "dur": 100,
                "args": {"Input Dims": [], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "123",
                "ts": 350, "dur": 40,
                "args": {"Input Dims": [], "External id": 4}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        self.assertEqual(len(profile.steps_costs), 2)
        step = profile.steps_costs[0]
        self.assertEqual(step.costs[ProfileRole.Kernel], 0)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 0)
        self.assertEqual(step.costs[ProfileRole.Memset], 0)
        self.assertEqual(step.costs[ProfileRole.Runtime], 0)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 0)
        self.assertEqual(step.costs[ProfileRole.CpuOp], 10)
        self.assertEqual(step.costs[ProfileRole.Other], 200 - 10)
        self.assertEqual(step.costs[ProfileRole.Total], 200)
        step = profile.steps_costs[1]
        self.assertEqual(step.costs[ProfileRole.Kernel], 0)
        self.assertEqual(step.costs[ProfileRole.Memcpy], 0)
        self.assertEqual(step.costs[ProfileRole.Memset], 0)
        self.assertEqual(step.costs[ProfileRole.Runtime], 0)
        self.assertEqual(step.costs[ProfileRole.DataLoader], 0)
        self.assertEqual(step.costs[ProfileRole.CpuOp], 40)
        self.assertEqual(step.costs[ProfileRole.Other], 100 - 40)
        self.assertEqual(step.costs[ProfileRole.Total], 100)

    # Test GPU utilization, est. SM efficiency, and occupancy.
    def test_gpu_utilization(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "aten::mat_mul", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 70,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 1, "tid": "stream 7",
                "ts": 130, "dur": 10,
                "args": {"correlation": 334, "external id": 4, "device": 1,
                        "blocks per SM": 0.5, "est. achieved occupancy %": 0.6}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 0,
                "args": {"correlation": 334, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void gemmSN_TN_kernel_64addr", "pid": 1, "tid": "stream 8",
                "ts": 135, "dur": 15,
                "args": {"correlation": 335, "external id": 2, "device": 1,
                        "blocks per SM": 0.6, "est. achieved occupancy %": 0.1}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void gemmSN_TN_kernel_64addr", "pid": 1, "tid": "stream 8",
                "ts": 150, "dur": 0,
                "args": {"correlation": 335, "external id": 2, "device": 1,
                        "blocks per SM": 0.3, "est. achieved occupancy %": 0.2}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 0,
                "args": {"correlation": 335, "external id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 1, "tid": "stream 7",
                "ts": 145, "dur": 25,
                "args": {"correlation": 336, "external id": 4, "device": 1,
                        "blocks per SM": 0.3, "est. achieved occupancy %": 1.0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 125, "dur": 3,
                "args": {"correlation": 336, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 1, "tid": "stream 7",
                "ts": 200, "dur": 20,
                "args": {"correlation": 337, "external id": 2, "device": 1,
                        "blocks per SM": 10.5, "est. achieved occupancy %": 0.3}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 195, "dur": 1,
                "args": {"correlation": 337, "external id": 2}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        self.assertEqual(len(profile.gpu_metrics_parser.gpu_ids), 1)
        self.assertAlmostEqual(profile.gpu_metrics_parser.gpu_utilization[1], (40 + 20) / 120)
        self.assertAlmostEqual(profile.gpu_metrics_parser.avg_approximated_sm_efficiency_per_device[1],
                               (0.5 * (135 - 130)
                                + 1.0 * (140 - 135)
                                + 0.6 * (145 - 140)
                                + 0.9 * (150 - 145)
                                + 0.3 * (170 - 150)
                                + 1.0 * (220 - 200)) / (220 - 100))
        self.assertAlmostEqual(profile.gpu_metrics_parser.avg_occupancy_per_device[1],
                               (0.6 * 10 + 0.1 * 15 + 1.0 * 25 + 0.3 * 20) / (10 + 15 + 25 + 20))

        gpu_util_expected = [(100, 0), (110, 0), (120, 0), (130, 1.0), (140, 1.0), (150, 1.0), (160, 1.0),
                             (170, 0), (180, 0), (190, 0), (200, 1.0), (210, 1.0), (220, 0)]
        for gpu_id in profile.gpu_metrics_parser.gpu_ids:
            buckets = profile.gpu_metrics_parser.gpu_util_buckets[gpu_id]
            gpu_util_id = 0
            for b in buckets:
                self.assertEqual(b[0], gpu_util_expected[gpu_util_id][0])
                self.assertAlmostEqual(b[1], gpu_util_expected[gpu_util_id][1])
                gpu_util_id += 1
            self.assertEqual(gpu_util_id, len(gpu_util_expected))

        sm_efficiency_expected = [(130, 0.5), (135, 0), (135, 1.0), (140, 0), (140, 0.6), (145, 0), (145, 0.9),
                                  (150, 0), (150, 0.3), (170, 0), (170, 0), (200, 0), (200, 1.0), (220, 0)]
        for gpu_id in profile.gpu_metrics_parser.gpu_ids:
            ranges = profile.gpu_metrics_parser.approximated_sm_efficiency_ranges[gpu_id]
            sm_efficiency_id = 0
            for r in ranges:
                self.assertEqual(
                    r[0], sm_efficiency_expected[sm_efficiency_id][0])
                self.assertAlmostEqual(
                    r[2], sm_efficiency_expected[sm_efficiency_id][1])
                sm_efficiency_id += 1
                self.assertEqual(
                    r[1], sm_efficiency_expected[sm_efficiency_id][0])
                self.assertAlmostEqual(
                    0, sm_efficiency_expected[sm_efficiency_id][1])
                sm_efficiency_id += 1
            self.assertEqual(sm_efficiency_id, len(sm_efficiency_expected))

        count = 0
        for agg_by_op in profile.kernel_list_groupby_name_op:
            if agg_by_op.name == 'void gemmSN_TN_kernel_64addr' and agg_by_op.op_name == 'aten::mat_mul':
                self.assertAlmostEqual(agg_by_op.avg_blocks_per_sm, 0.6)
                self.assertAlmostEqual(agg_by_op.avg_occupancy, 0.1)
                count += 1
            if agg_by_op.name == 'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>' and \
                    agg_by_op.op_name == 'aten::mm':
                self.assertAlmostEqual(
                    agg_by_op.avg_blocks_per_sm, (0.5 * 10 + 0.3 * 25) / (10 + 25))
                self.assertAlmostEqual(
                    agg_by_op.avg_occupancy, (0.6 * 10 + 1.0 * 25) / (10 + 25))
                count += 1
            if agg_by_op.name == 'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>' and \
                    agg_by_op.op_name == 'aten::mat_mul':
                self.assertAlmostEqual(agg_by_op.avg_blocks_per_sm, 10.5)
                self.assertAlmostEqual(agg_by_op.avg_occupancy, 0.3)
                count += 1
        self.assertEqual(count, 3)

        count = 0
        for _id, (name, row) in enumerate(profile.kernel_stat.iterrows()):
            # The kernel with zero 'dur' should be ignored.
            if name == 'void gemmSN_TN_kernel_64addr':
                self.assertAlmostEqual(row['blocks_per_sm'], 0.6)
                self.assertAlmostEqual(row['occupancy'], 0.1)
                count += 1
            if name == 'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>':
                self.assertAlmostEqual(
                    row['blocks_per_sm'], (0.5 * 10 + 0.3 * 25 + 10.5 * 20) / (10 + 25 + 20))
                self.assertAlmostEqual(
                    row['occupancy'], (0.6 * 10 + 1.0 * 25 + 0.3 * 20) / (10 + 25 + 20))
                count += 1
        self.assertEqual(count, 2)

    # Test GPU utilization 3 metrics works fine if kernel out of ProfilerStep.
    def test_gpu_utilization_kernel_out_of_step(self):
        json_content = """
            [{
                "ph": "X", "cat": "Operator",
                "name": "aten::mat_mul", "pid": 13721, "tid": "456",
                "ts": 10, "dur": 10,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 120, "dur": 70,
                "args": {"Input Dims": [], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::mm", "pid": 13721, "tid": "456",
                "ts": 220, "dur": 20,
                "args": {"Input Dims": [], "External id": 4}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#2", "pid": 13721, "tid": "456",
                "ts": 100, "dur": 100,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 1, "tid": "stream 7",
                "ts": 60, "dur": 20,
                "args": {"correlation": 334, "external id": 1, "device": 1,
                        "blocks per SM": 0.5, "est. achieved occupancy %": 0.6}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 15, "dur": 5,
                "args": {"correlation": 334, "external id": 1}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 1, "tid": "stream 7",
                "ts": 240, "dur": 25,
                "args": {"correlation": 337, "external id": 4, "device": 1,
                        "blocks per SM": 10.5, "est. achieved occupancy %": 0.3}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 230, "dur": 10,
                "args": {"correlation": 337, "external id": 4}
            }]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        self.assertEqual(len(profile.gpu_metrics_parser.gpu_ids), 1)
        self.assertAlmostEqual(profile.gpu_metrics_parser.gpu_utilization[1], 0.0)
        self.assertTrue(profile.gpu_metrics_parser.avg_approximated_sm_efficiency_per_device[1] is None)
        self.assertTrue(profile.gpu_metrics_parser.avg_occupancy_per_device[1] is None)
        self.assertTrue(profile.gpu_metrics_parser.blocks_per_sm_count[1] > 0)
        self.assertTrue(profile.gpu_metrics_parser.occupancy_count[1] > 0)

        count = 0
        for agg_by_op in profile.kernel_list_groupby_name_op:
            if agg_by_op.name == 'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>' \
                    and agg_by_op.op_name == 'aten::mat_mul':
                self.assertAlmostEqual(agg_by_op.avg_blocks_per_sm, 0.5)
                self.assertAlmostEqual(agg_by_op.avg_occupancy, 0.6)
                count += 1
            if agg_by_op.name == 'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>' and \
                    agg_by_op.op_name == 'aten::mm':
                self.assertAlmostEqual(
                    agg_by_op.avg_blocks_per_sm, 10.5)
                self.assertAlmostEqual(
                    agg_by_op.avg_occupancy, 0.3)
                count += 1
        self.assertEqual(count, 2)

        count = 0
        for _id, (name, row) in enumerate(profile.kernel_stat.iterrows()):
            # The kernel with zero 'dur' should be ignored.
            if name == 'void cunn_ClassNLLCriterion_updateGradInput_kernel<float>':
                self.assertAlmostEqual(row['blocks_per_sm'], (20 * 0.5 + 25 * 10.5) / (20 + 25))
                self.assertAlmostEqual(row['occupancy'], (20 * 0.6 + 25 * 0.3) / (20 + 25))
                count += 1
        self.assertEqual(count, 1)

    def test_dump_gpu_metrics(self):
        profile = RunProfile('test_dump_gpu_metrics', None)
        # Faked data for easy to see in UI. Real data values are 1/100 of these.
        gpu_util_buckets = [[(1621401187223005, 0.0), (1621401187224005, 0.0),
                            (1621401187225005, 0.6), (1621401187226005, 0.5),
                            (1621401187227005, 0.6), (1621401187228005, 0.2),
                            (1621401187229005, 0.6), (1621401187230005, 0.1),
                            (1621401187231005, 0.5), (1621401187232005, 0.2),
                            (1621401187233005, 0.3), (1621401187234005, 0.4),
                            (1621401187235005, 0.4219409282700422),
                            (1621401187236901, 0)]]
        # Faked data for easy to see in UI. Real data values are 1/10 of these.
        approximated_sm_efficiency_ranges = \
            [[(1621401187225275, 1621401187225278, 0.25), (1621401187225530, 1621401187225532, 0.125),
              (1621401187225820, 1621401187225821, 0.125), (1621401187226325, 1621401187226327, 0.25),
              (1621401187226575, 1621401187226577, 0.125), (1621401187226912, 1621401187226913, 0.125),
              (1621401187227092, 1621401187227094, 0.125), (1621401187227619, 1621401187227620, 0.125),
              (1621401187227745, 1621401187227746, 0.125), (1621401187227859, 1621401187227860, 0.125),
              (1621401187227973, 1621401187227974, 0.125), (1621401187228279, 1621401187228280, 0.125),
              (1621401187228962, 1621401187228963, 0.125), (1621401187229153, 1621401187229155, 0.125),
              (1621401187229711, 1621401187229715, 0.125), (1621401187230162, 1621401187230163, 0.125),
              (1621401187231100, 1621401187231103, 0.125), (1621401187231692, 1621401187231694, 0.5),
              (1621401187232603, 1621401187232604, 0.125), (1621401187232921, 1621401187232922, 0.125),
              (1621401187233342, 1621401187233343, 0.125), (1621401187233770, 1621401187233772, 0.125),
              (1621401187234156, 1621401187234159, 0.125), (1621401187234445, 1621401187234446, 0.125),
              (1621401187235025, 1621401187235028, 0.125), (1621401187235555, 1621401187235556, 0.125),
              (1621401187236158, 1621401187236159, 0.125), (1621401187236278, 1621401187236279, 0.125),
              (1621401187236390, 1621401187236391, 0.125), (1621401187236501, 1621401187236502, 0.125)]]

        basedir = os.path.dirname(os.path.realpath(__file__))
        trace_json_flat_path = os.path.join(basedir, 'gpu_metrics_input.json')
        gpu_metrics_parser = GPUMetricsParser()
        gpu_metrics_parser.gpu_util_buckets = gpu_util_buckets
        gpu_metrics_parser.approximated_sm_efficiency_ranges = approximated_sm_efficiency_ranges
        profile.gpu_metrics = gpu_metrics_parser.get_gpu_metrics()
        with open(trace_json_flat_path, 'rb') as file:
            raw_data = file.read()
        data_with_gpu_metrics_compressed = profile.append_gpu_metrics(raw_data)
        data_with_gpu_metrics_flat = gzip.decompress(
            data_with_gpu_metrics_compressed)

        trace_json_expected_path = os.path.join(basedir, 'gpu_metrics_expected.json')
        with open(trace_json_expected_path, 'rb') as file:
            data_expected = file.read()

        # Parse to json in order to ignore text format difference.
        data_with_gpu_metrics_json = json.loads(
            data_with_gpu_metrics_flat.decode('utf8'))
        data_expected_json = json.loads(data_expected.decode('utf8'))
        data_with_gpu_metrics_str = json.dumps(
            data_with_gpu_metrics_json, sort_keys=True)
        data_expected_str = json.dumps(data_expected_json, sort_keys=True)

        self.assertEqual(data_with_gpu_metrics_str, data_expected_str)

        try:
            _ = json.loads(data_with_gpu_metrics_flat.decode('utf8'))
        except Exception:
            self.assertTrue(
                False, 'The string fails to be parsed by json after appending gpu metrics.')

    def test_memory_view(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 10, "dur": 10,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "enumerate(DataLoader)#_SingleProcessDataLoaderIter.__next__", "pid": 13721, "tid": "123",
                "ts": 100, "dur": 180,
                "args": {"Input Dims": [], "External id": 2}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 200, "dur": 60,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::nll_loss_backward", "pid": 13721, "tid": "123",
                "ts": 340, "dur": 70,
                "args": {"Input Dims": [[], [32, 1000], [32], [], [], [], []], "External id": 4}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#1", "pid": 13721, "tid": "123",
                "ts": 50, "dur": 400,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "ProfilerStep#2", "pid": 13721, "tid": "123",
                "ts": 500, "dur": 500,
                "args": {"Input Dims": [], "External id": 1}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 510, "dur": 150,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::copy_", "pid": 13721, "tid": "123",
                "ts": 520, "dur": 100,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },

            {
                "ph": "X", "cat": "Operator",
                "name": "aten::liner", "pid": 13721, "tid": "123",
                "ts": 700, "dur": 100,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::t", "pid": 13721, "tid": "123",
                "ts": 705, "dur": 40,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::transpose", "pid": 13721, "tid": "123",
                "ts": 710, "dur": 30,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::tranas_stride", "pid": 13721, "tid": "123",
                "ts": 720, "dur": 10,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::addmm", "pid": 13721, "tid": "123",
                "ts": 750, "dur": 40,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::to", "pid": 13721, "tid": "123",
                "ts": 900, "dur": 100,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Memcpy",
                "name": "Memcpy HtoD (Pageable -> Device)", "pid": 0, "tid": "stream 7",
                "ts": 405, "dur": 10,
                "args": {"stream": 7, "correlation": 334, "external id": 4}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaMemcpyAsync", "pid": 13721, "tid": "456",
                "ts": 360, "dur": 20,
                "args": {"correlation": 334, "external id": 4}
            },
            {
                "ph": "X", "cat": "Memset",
                "name": "Memset (Device)", "pid": 0, "tid": "stream 7",
                "ts": 420, "dur": 5,
                "args": {"stream": 7, "correlation": 40344, "external id": 4}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaMemsetAsync", "pid": 13721, "tid": "456",
                "ts": 390, "dur": 10,
                "args": {"correlation": 40344, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "void cunn_ClassNLLCriterion_updateGradInput_kernel<float>", "pid": 0, "tid": "stream 7",
                "ts": 430, "dur": 15,
                "args": {"correlation": 40348, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 405, "dur": 5,
                "args": {"correlation": 40348, "external id": 4}
            },


            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 90,
                "args": {
                "Device Type": 0, "Device Id": -1, "Addr": 90, "Bytes": 4
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 150,
                "args": {
                "Device Type": 0, "Device Id": -1, "Addr": 150, "Bytes": 4
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 200,
                "args": {
                "Device Type": 0, "Device Id": -1, "Addr": 200, "Bytes": 4
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 210,
                "args": {
                "Device Type": 1, "Device Id": 0, "Addr": 210, "Bytes": 4
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 265,
                "args": {
                "Device Type": 1, "Device Id": 0, "Addr": 265, "Bytes": 4
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 300,
                "args": {
                "Device Type": 1, "Device Id": 0, "Addr": 300, "Bytes": 4
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 350,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 350, "Bytes": 10
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 360,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 350, "Bytes": -10
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 450,
                "args": {
                    "Device Type": 0, "Device Id": -1, "Addr": 450, "Bytes": 1000000
                }
            },

            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 515,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 515, "Bytes": 100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 520,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 520, "Bytes": 100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 600,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 520, "Bytes": -100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 690,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 690, "Bytes": 100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 701,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 701, "Bytes": 100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 796,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 515, "Bytes": -100
                }
            },

            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 708,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 708, "Bytes": 100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 742,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 708, "Bytes": -100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 715,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 715, "Bytes": 50
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 735,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 715, "Bytes": -50
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 725,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 725, "Bytes": 50
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 728,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 725, "Bytes": -50
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 729,
                "args": {
                    "Device Type": 0, "Device Id": -1, "Addr": 729, "Bytes": 50
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 746,
                "args": {
                    "Device Type": 0, "Device Id": -1, "Addr": 746, "Bytes": 100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 747,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 747, "Bytes": 20
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 749,
                "args": {
                    "Device Type": 0, "Device Id": -1, "Addr": 690, "Bytes": -100
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 760,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 760, "Bytes": 30
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 780,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 760, "Bytes": -30
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 795,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 795, "Bytes": 10
                }
            },
            {
                "ph": "i", "s": "t", "name": "[memory]",
                "pid": 13721, "tid": 123,
                "ts": 799,
                "args": {
                    "Device Type": 1, "Device Id": 0, "Addr": 795, "Bytes": -10
                }
            }
        ]
        """
        import logging

        from torch_tb_profiler.utils import get_logger
        logger = get_logger()
        logger.addHandler(logging.StreamHandler())

        profile = parse_json_trace(json_content)
        profile.process()
        memory_stats = profile.memory_snapshot.get_memory_statistics(profile.tid2tree)

        self.assertEqual(len(memory_stats), 2)
        self.assertIn('GPU0', memory_stats)

        # validation
        gpu_expected_data = {
            # self increase size, self allocation size, self allocation count, increase size, allocation size, allocation count, call # noqa: E501
            'aten::to': [104, 104, 2, 104, 204, 3, 4],
            'aten::nll_loss_backward': [0, 10, 1, 0, 10, 1, 1],
            'aten::copy_': [0, 100, 1, 0, 100, 1, 1],
            'aten::addmm': [0, 30, 1, 0, 30, 1, 1],
            'aten::tranas_stride': [0, 50, 1, 0, 50, 1, 1],
            'aten::transpose': [0, 50, 1, 0, 100, 2, 1],
            'aten::t': [0, 100, 1, 0, 200, 3, 1],
            'aten::liner': [20, 130, 3, 20, 360, 7, 1]
        }

        cpu_expected_data = {
            'aten::to': [4, 4, 1, 4, 4, 1, 4],
            'aten::liner': [0, 100, 1, 50, 150, 2, 1],
            'aten::tranas_stride': [50, 50, 1, 50, 50, 1, 1],
            'aten::transpose': [0, 0, 0, 50, 50, 1, 1],
            'aten::t': [0, 0, 0, 50, 50, 1, 1]
        }

        validate_data = [
            (memory_stats['CPU'], cpu_expected_data),
            (memory_stats['GPU0'], gpu_expected_data)
        ]
        for (mem_stat, expected_data) in validate_data:
            for name, values in expected_data.items():
                self.assertEqual(mem_stat[name], values)

    # Test group by 'kernel detail + op name'.
    def test_group_by_kernel_columns(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "op1", "pid": 13721, "tid": "123",
                "ts": 200, "dur": 60,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "op2", "pid": 13721, "tid": "456",
                "ts": 340, "dur": 70,
                "args": {"Input Dims": [[], [32, 1000], [32], [], [], [], []], "External id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 230, "dur": 15,
                "args": {"correlation": 1000, "external id": 3, "device": 0,
                        "grid": [16, 1, 1], "block": [16, 16, 16], "registers per thread": 18, "shared memory": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 210, "dur": 5,
                "args": {"correlation": 1000, "external id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 250, "dur": 10,
                "args": {"correlation": 1001, "external id": 3, "device": 0,
                        "grid": [16, 1, 1], "block": [16, 16, 16], "registers per thread": 18, "shared memory": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 215, "dur": 5,
                "args": {"correlation": 1001, "external id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 250, "dur": 13,
                "args": {"correlation": 1002, "external id": 3, "device": 0,
                        "grid": [16, 1, 1], "block": [16, 16, 64], "registers per thread": 18, "shared memory": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 220, "dur": 5,
                "args": {"correlation": 1002, "external id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 250, "dur": 17,
                "args": {"correlation": 1003, "external id": 4, "device": 0,
                        "grid": [16, 1, 1], "block": [16, 16, 64], "registers per thread": 18, "shared memory": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 350, "dur": 5,
                "args": {"correlation": 1003, "external id": 4}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        expected_agg_kernels = [
            {
                'name': 'kernel1',
                'op_name': 'op1',
                'grid': '[16, 1, 1]',
                'block': '[16, 16, 16]',
                'registers per thread': 18,
                'shared memory': 0,
                'calls': 2,
                'total_duration': 15 + 10,
                'avg_duration': (15 + 10) / 2,
                'min_duration': min(15, 10),
                'max_duration': max(15, 10)
            },
            {
                'name': 'kernel1',
                'op_name': 'op1',
                'grid': '[16, 1, 1]',
                'block': '[16, 16, 64]',  # Only changed this.
                'registers per thread': 18,
                'shared memory': 0,
                'calls': 1,
                'total_duration': 13,
                'avg_duration': 13,
                'min_duration': 13,
                'max_duration': 13
            },
            {
                'name': 'kernel1',
                'op_name': 'op2',  # Only changed this.
                'grid': '[16, 1, 1]',
                'block': '[16, 16, 64]',
                'registers per thread': 18,
                'shared memory': 0,
                'calls': 1,
                'total_duration': 17,
                'avg_duration': 17,
                'min_duration': 17,
                'max_duration': 17
            }
        ]
        index = 0
        self.assertEqual(len(profile.kernel_list_groupby_name_op), len(expected_agg_kernels))
        for agg_kernel in profile.kernel_list_groupby_name_op:
            expected_agg_kernel = expected_agg_kernels[index]
            self.assertEqual(agg_kernel.name, expected_agg_kernel['name'])
            self.assertEqual(agg_kernel.op_name, expected_agg_kernel['op_name'])
            self.assertEqual(str(agg_kernel.grid), expected_agg_kernel['grid'])
            self.assertEqual(str(agg_kernel.block), expected_agg_kernel['block'])
            self.assertEqual(agg_kernel.regs_per_thread, expected_agg_kernel['registers per thread'])
            self.assertEqual(agg_kernel.shared_memory, expected_agg_kernel['shared memory'])
            self.assertEqual(agg_kernel.calls, expected_agg_kernel['calls'])
            self.assertEqual(agg_kernel.total_duration, expected_agg_kernel['total_duration'])
            self.assertAlmostEqual(agg_kernel.avg_duration, expected_agg_kernel['avg_duration'])
            self.assertEqual(agg_kernel.min_duration, expected_agg_kernel['min_duration'])
            self.assertEqual(agg_kernel.max_duration, expected_agg_kernel['max_duration'])
            index += 1

    # Test group by 'kernel detail + op name' with invalid input lack of some kernel field
    def test_group_by_kernel_columns_invalid_input(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "op1", "pid": 13721, "tid": "123",
                "ts": 200, "dur": 60,
                "args": {"Input Dims": [[2, 8, 5], [], [], [], [], [], [], []], "External id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 220, "dur": 1,
                "args": {"correlation": 1000, "external id": 3, "device": 0,
                        "block": [16, 16, 16], "registers per thread": 18, "shared memory": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 210, "dur": 5,
                "args": {"correlation": 1000, "external id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 230, "dur": 2,
                "args": {"correlation": 1001, "external id": 3, "device": 0,
                        "grid": [16, 1, 1], "registers per thread": 18, "shared memory": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 220, "dur": 5,
                "args": {"correlation": 1001, "external id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 240, "dur": 3,
                "args": {"correlation": 1002, "external id": 3, "device": 0,
                        "grid": [16, 1, 1], "block": [16, 16, 16], "shared memory": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 230, "dur": 5,
                "args": {"correlation": 1002, "external id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 250, "dur": 4,
                "args": {"correlation": 1003, "external id": 3, "device": 0,
                        "grid": [16, 1, 1], "block": [16, 16, 16], "registers per thread": 18}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 240, "dur": 5,
                "args": {"correlation": 1003, "external id": 3}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel1", "pid": 0, "tid": "stream 7",
                "ts": 260, "dur": 5,
                "args": {"correlation": 1004, "external id": 3, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "456",
                "ts": 250, "dur": 5,
                "args": {"correlation": 1004, "external id": 3}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()
        expected_agg_kernels = [
            {
                'name': 'kernel1',
                'op_name': 'op1',
                'grid': None,
                'block': [16, 16, 16],
                'registers per thread': 18,
                'shared memory': 0,
                'calls': 1,
                'total_duration': 1,
                'avg_duration': 1,
                'min_duration': 1,
                'max_duration': 1
            },
            {
                'name': 'kernel1',
                'op_name': 'op1',
                'grid': [16, 1, 1],
                'block': None,
                'registers per thread': 18,
                'shared memory': 0,
                'calls': 1,
                'total_duration': 2,
                'avg_duration': 2,
                'min_duration': 2,
                'max_duration': 2
            },
            {
                'name': 'kernel1',
                'op_name': 'op1',
                'grid': [16, 1, 1],
                'block': [16, 16, 16],
                'registers per thread': None,
                'shared memory': 0,
                'calls': 1,
                'total_duration': 3,
                'avg_duration': 3,
                'min_duration': 3,
                'max_duration': 3
            },
            {
                'name': 'kernel1',
                'op_name': 'op1',
                'grid': [16, 1, 1],
                'block': [16, 16, 16],
                'registers per thread': 18,
                'shared memory': None,
                'calls': 1,
                'total_duration': 4,
                'avg_duration': 4,
                'min_duration': 4,
                'max_duration': 4
            },
            {
                'name': 'kernel1',
                'op_name': 'op1',
                'grid': None,
                'block': None,
                'registers per thread': None,
                'shared memory': None,
                'calls': 1,
                'total_duration': 5,
                'avg_duration': 5,
                'min_duration': 5,
                'max_duration': 5
            }
        ]
        index = 0
        self.assertEqual(len(profile.kernel_list_groupby_name_op), len(expected_agg_kernels))
        for agg_kernel in profile.kernel_list_groupby_name_op:
            expected_agg_kernel = expected_agg_kernels[index]
            self.assertEqual(agg_kernel.name, expected_agg_kernel['name'])
            self.assertEqual(agg_kernel.op_name, expected_agg_kernel['op_name'])
            self.assertEqual(agg_kernel.grid, expected_agg_kernel['grid'])
            self.assertEqual(agg_kernel.block, expected_agg_kernel['block'])
            self.assertEqual(agg_kernel.regs_per_thread, expected_agg_kernel['registers per thread'])
            print(agg_kernel.name, agg_kernel.grid, agg_kernel.block, agg_kernel.shared_memory)
            self.assertEqual(agg_kernel.shared_memory, expected_agg_kernel['shared memory'])
            self.assertEqual(agg_kernel.calls, expected_agg_kernel['calls'])
            self.assertEqual(agg_kernel.total_duration, expected_agg_kernel['total_duration'])
            self.assertAlmostEqual(agg_kernel.avg_duration, expected_agg_kernel['avg_duration'])
            self.assertEqual(agg_kernel.min_duration, expected_agg_kernel['min_duration'])
            self.assertEqual(agg_kernel.max_duration, expected_agg_kernel['max_duration'])
            index += 1

    # Test tensor core related feature.
    def test_tensor_core(self):
        json_content = """[
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::conv2d", "pid": 13721, "tid": "123",
                "ts": 200, "dur": 100,
                "args": {"Input Dims": [[]], "External id": 3}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "op_no_tc", "pid": 13721, "tid": "123",
                "ts": 205, "dur": 10,
                "args": {"Input Dims": [[]], "External id": 4}
            },
            {
                "ph": "X", "cat": "Operator",
                "name": "aten::cudnn_convolution", "pid": 13721, "tid": "123",
                "ts": 215, "dur": 10,
                "args": {"Input Dims": [[]], "External id": 5}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "kernel_no_tc", "pid": 0, "tid": "stream 7",
                "ts": 210, "dur": 10,
                "args": {"correlation": 1000, "external id": 4, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "123",
                "ts": 205, "dur": 5,
                "args": {"correlation": 1000, "external id": 4}
            },
            {
                "ph": "X", "cat": "Kernel",
                "name": "volta_fp16_s884cudnn_fp16_128x128_ldg8_splitK_relu_f2f_exp_small_nhwc_tn_v1",
                "pid": 0, "tid": "stream 7",
                "ts": 220, "dur": 15,
                "args": {"correlation": 1001, "external id": 5, "device": 0}
            },
            {
                "ph": "X", "cat": "Runtime",
                "name": "cudaLaunchKernel", "pid": 13721, "tid": "123",
                "ts": 215, "dur": 5,
                "args": {"correlation": 1001, "external id": 5}
            }
        ]
        """
        profile = parse_json_trace(json_content)
        profile.process()

        expected_agg_ops = {
            'aten::conv2d': {
                'tc_eligible': True,
                'tc_self_ratio': 0,
                'tc_total_ratio': 15 / (15 + 10)
            },
            'op_no_tc': {
                'tc_eligible': False,
                'tc_self_ratio': 0,
                'tc_total_ratio': 0
            },
            'aten::cudnn_convolution': {
                'tc_eligible': True,
                'tc_self_ratio': 1.0,
                'tc_total_ratio': 1.0
            }
        }
        self.assertEqual(len(profile.op_list_groupby_name), len(expected_agg_ops))
        for agg_op in profile.op_list_groupby_name:
            expected_agg_op = expected_agg_ops[agg_op.name]
            self.assertEqual(agg_op.tc_eligible, expected_agg_op['tc_eligible'])
            self.assertAlmostEqual(agg_op.tc_self_ratio, expected_agg_op['tc_self_ratio'])
            self.assertAlmostEqual(agg_op.tc_total_ratio, expected_agg_op['tc_total_ratio'])

        expected_kernels_groupby_op = {
            'kernel_no_tc': {
                'op_name': 'op_no_tc',
                'tc_used': False,
                'op_tc_eligible': False
            },
            'volta_fp16_s884cudnn_fp16_128x128_ldg8_splitK_relu_f2f_exp_small_nhwc_tn_v1': {
                'op_name': 'aten::cudnn_convolution',
                'tc_used': True,
                'op_tc_eligible': True
            }
        }
        self.assertEqual(len(profile.kernel_list_groupby_name_op), len(expected_kernels_groupby_op))
        for agg_kernel in profile.kernel_list_groupby_name_op:
            expected_agg_kernel = expected_kernels_groupby_op[agg_kernel.name]
            self.assertEqual(agg_kernel.op_name, expected_agg_kernel['op_name'])
            self.assertEqual(agg_kernel.tc_used, expected_agg_kernel['tc_used'])
            self.assertEqual(agg_kernel.op_tc_eligible, expected_agg_kernel['op_tc_eligible'])

        self.assertAlmostEqual(profile.tc_ratio[0], 15 / (15 + 10))
        self.assertAlmostEqual(profile.tc_eligible_ops_kernel_ratio, 15 / (15 + 10))


class TestDistributed(unittest.TestCase):

    def test_distributed_nccl(self):
        json_content0 = """[
            {
            "ph": "X", "cat": "cpu_op",
            "name": "nccl:broadcast", "pid": 23803, "tid": "23803",
            "ts": 0, "dur": 75,
            "args": {"External id": 146, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "ncclKernel_Broadcast_RING_LL_Sum_int8_t(ncclWorkElem)", "pid": 0, "tid": "stream 16",
            "ts": 16, "dur": 16,
            "args": {"device": 0, "correlation": 28506, "external id": 146}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::add_", "pid": 23803, "tid": "23803",
            "ts": 100, "dur": 20,
            "args": {"External id": 24504, "Input Dims": [[1000], [1000], []], "Input type": ["float", "float", "Int"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "void at::native::vectorized_elementwise_kernel", "pid": 0, "tid": "stream 7",
            "ts": 130, "dur": 161,
            "args": {"device": 0, "correlation": 99765, "external id": 24504}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "nccl:all_reduce", "pid": 23803, "tid": "25166",
            "ts": 160, "dur": 75,
            "args": {"External id": 2513, "Input Dims": [[2049000]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "ncclKernel_AllReduce_RING_LL_Sum_float(ncclWorkElem)", "pid": 0, "tid": "stream 16",
            "ts": 162, "dur": 1556,
            "args": {"device": 0, "correlation": 33218, "external id": 2513}
            }
        ]
        """
        json_content1 = """[
            {
            "ph": "X", "cat": "cpu_op",
            "name": "nccl:broadcast", "pid": 23803, "tid": "23803",
            "ts": 0, "dur": 20,
            "args": {"External id": 146, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "ncclKernel_Broadcast_RING_LL_Sum_int8_t(ncclWorkElem)", "pid": 0, "tid": "stream 16",
            "ts": 8, "dur": 31,
            "args": {"device": 0, "correlation": 28506, "external id": 146}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::add_", "pid": 23803, "tid": "23803",
            "ts": 25, "dur": 20,
            "args": {"External id": 24504, "Input Dims": [[1000], [1000], []], "Input type": ["float", "float", "Int"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "void at::native::vectorized_elementwise_kernel", "pid": 0, "tid": "stream 7",
            "ts": 30, "dur": 161,
            "args": {"device": 0, "correlation": 99765, "external id": 24504}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "nccl:all_reduce", "pid": 23803, "tid": "25166",
            "ts": 160, "dur": 75,
            "args": {"External id": 2513, "Input Dims": [[2049000]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "ncclKernel_AllReduce_RING_LL_Sum_float(ncclWorkElem)", "pid": 0, "tid": "stream 16",
            "ts": 562, "dur": 1058,
            "args": {"device": 0, "correlation": 33218, "external id": 2513}
            }
        ]
        """

        profile0 = parse_json_trace(json_content0, 'worker0')
        dist_data0 = DistributedRunProfileData(profile0)
        self.assertTrue(profile0.has_communication)
        self.assertEqual(len(profile0.comm_node_list), 2)
        self.assertEqual(profile0.steps_costs[0].costs, [105, 0, 0, 16, 0, 0, 79, 35, 235])

        profile1 = parse_json_trace(json_content1, 'worker1')
        dist_data1 = DistributedRunProfileData(profile1)
        self.assertTrue(profile1.has_communication)
        self.assertEqual(len(profile1.comm_node_list), 2)
        self.assertEqual(profile1.steps_costs[0].costs[3], 22)

        loader = RunLoader('test_nccl', '', None)
        dist_profile = loader._process_distributed_profiles([dist_data0, dist_data1], 0)
        self.assertEqual(dist_profile.steps_to_overlap['data']['0']['worker0'], [32, 73, 16, 114])
        self.assertEqual(dist_profile.steps_to_overlap['data']['0']['worker1'], [152, 9, 22, 52])
        self.assertEqual(dist_profile.steps_to_wait['data']['0']['worker0'], [1074, 498])
        self.assertEqual(dist_profile.steps_to_wait['data']['0']['worker1'], [1074, 15])
        self.assertEqual(dist_profile.comm_ops['data']['worker0']['rows'],
                         [['nccl:broadcast', 1, 212480, 212480, 16, 16, 16, 16],
                          ['nccl:all_reduce', 1, 8196000, 8196000, 1556, 1556, 1058, 1058]])
        self.assertEqual(dist_profile.comm_ops['data']['worker1']['rows'],
                         [['nccl:broadcast', 1, 212480, 212480, 31, 31, 16, 16],
                          ['nccl:all_reduce', 1, 8196000, 8196000, 1058, 1058, 1058, 1058]])

    def test_distributed_gloo_gpu(self):
        json_content0 = """[
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 16, "dur": 38,
            "args": {"External id": 165, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23805",
            "ts": 25, "dur": 36,
            "args": {"External id": 166, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 66, "dur": 18,
            "args": {"External id": 167, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::add_", "pid": 23803, "tid": "23800",
            "ts": 0, "dur": 20,
            "args": {"External id": 24504, "Input Dims": [[1000], [1000], []], "Input type": ["float", "float", "Int"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "void at::native::vectorized_elementwise_kernel", "pid": 0, "tid": "stream 7",
            "ts": 30, "dur": 101,
            "args": {"device": 0, "correlation": 99765, "external id": 24504}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23805",
            "ts": 110, "dur": 18,
            "args": {"External id": 2513, "Input Dims": [[2049000]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23803",
            "ts": 120, "dur": 36,
            "args": {"External id": 2516, "Input Dims": [[2049000]], "Input type": ["float"]}
            }
        ]
        """
        json_content1 = """[
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 20, "dur": 28,
            "args": {"External id": 256, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23805",
            "ts": 28, "dur": 30,
            "args": {"External id": 257, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 77, "dur": 6,
            "args": {"External id": 258, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::add_", "pid": 23803, "tid": "23800",
            "ts": 0, "dur": 30,
            "args": {"External id": 24504, "Input Dims": [[1000], [1000], []], "Input type": ["float", "float", "Int"]}
            },
            {
            "ph": "X", "cat": "Kernel",
            "name": "void at::native::vectorized_elementwise_kernel", "pid": 0, "tid": "stream 7",
            "ts": 70, "dur": 70,
            "args": {"device": 0, "correlation": 99765, "external id": 24504}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23805",
            "ts": 88, "dur": 38,
            "args": {"External id": 2513, "Input Dims": [[2049000]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23803",
            "ts": 130, "dur": 16,
            "args": {"External id": 2516, "Input Dims": [[2049000]], "Input type": ["float"]}
            }
        ]
        """

        profile0 = parse_json_trace(json_content0, 'worker0')
        dist_data0 = DistributedRunProfileData(profile0)
        self.assertTrue(profile0.has_communication)
        self.assertEqual(len(profile0.comm_node_list), 5)
        self.assertEqual(profile0.steps_costs[0].costs, [101, 0, 0, 39, 0, 0, 16, 0, 156])

        profile1 = parse_json_trace(json_content1, 'worker1')
        dist_data1 = DistributedRunProfileData(profile1)
        self.assertTrue(profile1.has_communication)
        self.assertEqual(len(profile1.comm_node_list), 5)
        self.assertEqual(profile1.steps_costs[0].costs, [70, 0, 0, 44, 0, 0, 20, 12, 146])

        loader = RunLoader('test_gloo_gpu', '', None)
        dist_profile = loader._process_distributed_profiles([dist_data0, dist_data1], 0)
        self.assertEqual(dist_profile.steps_to_overlap['data']['0']['worker0'], [31, 70, 39, 16])
        self.assertEqual(dist_profile.steps_to_overlap['data']['0']['worker1'], [16, 54, 44, 32])
        self.assertEqual(dist_profile.steps_to_wait['data']['0']['worker0'], [75, 34])
        self.assertEqual(dist_profile.steps_to_wait['data']['0']['worker1'], [78, 20])
        self.assertEqual(dist_profile.comm_ops['data']['worker0']['rows'],
                         [['gloo:broadcast', 3, 637440, 212480, 63, 21, 41, 14],
                          ['gloo:all_reduce', 2, 16392000, 8196000, 46, 23, 34, 17]])
        self.assertEqual(dist_profile.comm_ops['data']['worker1']['rows'],
                         [['gloo:broadcast', 3, 637440, 212480, 44, 15, 44, 15],
                          ['gloo:all_reduce', 2, 16392000, 8196000, 54, 27, 34, 17]])

    def test_distributed_gloo_cpu(self):
        json_content0 = """[
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 16, "dur": 38,
            "args": {"External id": 165, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23805",
            "ts": 25, "dur": 36,
            "args": {"External id": 166, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 66, "dur": 18,
            "args": {"External id": 167, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::add_", "pid": 23803, "tid": "23800",
            "ts": 0, "dur": 20,
            "args": {"External id": 24504, "Input Dims": [[1000], [1000], []], "Input type": ["float", "float", "Int"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::mul", "pid": 23803, "tid": "23800",
            "ts": 30, "dur": 101,
            "args": {"External id": 24505}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23805",
            "ts": 110, "dur": 18,
            "args": {"External id": 2513, "Input Dims": [[2049000]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23803",
            "ts": 120, "dur": 36,
            "args": {"External id": 2516, "Input Dims": [[2049000]], "Input type": ["float"]}
            }
        ]
        """
        json_content1 = """[
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 20, "dur": 28,
            "args": {"External id": 256, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23805",
            "ts": 28, "dur": 30,
            "args": {"External id": 257, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:broadcast", "pid": 23803, "tid": "23803",
            "ts": 77, "dur": 6,
            "args": {"External id": 258, "Input Dims": [[53120]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::add_", "pid": 23803, "tid": "23800",
            "ts": 0, "dur": 30,
            "args": {"External id": 24504, "Input Dims": [[1000], [1000], []], "Input type": ["float", "float", "Int"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "aten::mul", "pid": 23803, "tid": "23800",
            "ts": 70, "dur": 70,
            "args": {"External id": 24505}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23805",
            "ts": 88, "dur": 38,
            "args": {"External id": 2513, "Input Dims": [[2049000]], "Input type": ["float"]}
            },
            {
            "ph": "X", "cat": "cpu_op",
            "name": "gloo:all_reduce", "pid": 23803, "tid": "23803",
            "ts": 130, "dur": 16,
            "args": {"External id": 2516, "Input Dims": [[2049000]], "Input type": ["float"]}
            }
        ]
        """

        profile0 = parse_json_trace(json_content0, 'worker0')
        dist_data0 = DistributedRunProfileData(profile0)
        self.assertTrue(profile0.has_communication)
        self.assertEqual(len(profile0.comm_node_list), 5)
        self.assertEqual(profile0.steps_costs[0].costs, [0, 0, 0, 109, 0, 0, 47, 0, 156])

        profile1 = parse_json_trace(json_content1, 'worker1')
        dist_data1 = DistributedRunProfileData(profile1)
        self.assertTrue(profile1.has_communication)
        self.assertEqual(len(profile1.comm_node_list), 5)
        self.assertEqual(profile1.steps_costs[0].costs, [0, 0, 0, 98, 0, 0, 36, 12, 146])

        loader = RunLoader('test_gloo_cpu', '', None)
        dist_profile = loader._process_distributed_profiles([dist_data0, dist_data1], 0)
        self.assertEqual(dist_profile.steps_to_overlap['data']['0']['worker0'], [47, 74, 35, 0])
        self.assertEqual(dist_profile.steps_to_overlap['data']['0']['worker1'], [36, 64, 34, 12])
        self.assertEqual(dist_profile.steps_to_wait['data']['0']['worker0'], [75, 34])
        self.assertEqual(dist_profile.steps_to_wait['data']['0']['worker1'], [78, 20])
        self.assertEqual(dist_profile.comm_ops['data']['worker0']['rows'],
                         [['gloo:broadcast', 3, 637440, 212480, 63, 21, 41, 14],
                          ['gloo:all_reduce', 2, 16392000, 8196000, 46, 23, 34, 17]])
        self.assertEqual(dist_profile.comm_ops['data']['worker1']['rows'],
                         [['gloo:broadcast', 3, 637440, 212480, 44, 15, 44, 15],
                          ['gloo:all_reduce', 2, 16392000, 8196000, 54, 27, 34, 17]])


class TestMemoryCurve(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.event_data_cpu = [
            [1, 0, 0, 1, 4, 4, 0],          # alloc 1
            [20, 0, 0, 1, -4, 0, 0],        # free  1
            [100, 0, 0, 2, 8000, 8000, 0],  # alloc 2
            [200, 0, 0, 2, -8000, 0, 0],    # free  2
            [300, 0, 0, 3, 4, 4, 0],        # alloc 3
            [400, 0, 0, 4, 16, 20, 0],      # alloc 4
            [500, 0, 0, 5, 4000, 4020, 0],  # alloc 5
            [600, 0, 0, 4, -16, 4004, 0],   # free  4
            [700, 0, 0, 7, 80, 4084, 0],    # alloc 7
            [800, 0, 0, 3, -4, 4080, 0],    # free  3
            [900, 0, 0, 7, -80, 4000, 0],   # free  7
            [905, 0, 0, 4, -4000, 0, 0],    # free  5
        ]

        self.event_data_gpu = [
            [2, 1, 0, 11, 400, 400, 512],         # alloc 11
            [22, 1, 0, 11, -400, 0, 512],         # free  11
            [105, 1, 0, 12, 5000, 5000, 10240],   # alloc 12
            [106, 1, 0, 13, 3000, 8000, 10240],   # alloc 13
            [205, 1, 0, 12, -5000, 3000, 10240],  # free  12
            [401, 1, 0, 14, 1024, 4024, 10240],   # alloc 14
            [499, 1, 0, 15, 4, 4028, 10240],      # alloc 15
            [501, 1, 0, 13, -3000, 1028, 10240],  # free  13
            [502, 1, 0, 15, -4, 1024, 10240],     # free  15
            [906, 1, 0, 14, -1024, 0, 10240],     # free  14
        ]

        self.all_events = sorted(self.event_data_cpu + self.event_data_gpu, key=lambda e: e[0])

    def entry(self, ts, dev, dev_id, addr, alloc_size, total_allocated, total_reserved):
        return {
            'ph': 'i', 's': 't', 'name': '[memory]', 'pid': 0, 'tid': 0, 'ts': ts,
            'args': {
                'Device Type': dev,
                'Device Id': dev_id,
                'Addr': addr,
                'Bytes': alloc_size,
                'Total Allocated': total_allocated,
                'Total Reserved': total_reserved,
            },
        }

    def test_memory_curve_no_step_plot(self):
        json_content = json.dumps([self.entry(*data) for data in self.all_events])

        profile = parse_json_trace(json_content)
        profile.process()
        result = RunProfile.get_memory_curve(profile, time_metric='us', memory_metric='B', patch_for_step_plot=False)

        start_ts = profile.profiler_start_ts
        self.assertEqual(1, start_ts)

        curves = result['rows']

        self.assertIn('CPU', curves)
        self.assertIn('GPU0', curves)

        self.assertEqual(len(self.event_data_cpu), len(curves['CPU']))
        for i in range(len(self.event_data_cpu)):
            # adjusted timestamp
            self.assertEqual(self.event_data_cpu[i][0] - start_ts,  curves['CPU'][i][0])
            # total allocated
            self.assertEqual(self.event_data_cpu[i][-2], curves['CPU'][i][1])
            # total reserved
            self.assertEqual(self.event_data_cpu[i][-1], curves['CPU'][i][2])

        self.assertEqual(len(self.event_data_gpu), len(curves['GPU0']))
        for i in range(len(self.event_data_gpu)):
            self.assertEqual(self.event_data_gpu[i][0] - start_ts,  curves['GPU0'][i][0])
            self.assertEqual(self.event_data_gpu[i][-2], curves['GPU0'][i][1])
            self.assertEqual(self.event_data_gpu[i][-1], curves['GPU0'][i][2])

    def test_memory_curve_step_plot(self):
        json_content = json.dumps([self.entry(*data) for data in self.all_events])

        profile = parse_json_trace(json_content)
        profile.process()
        result = RunProfile.get_memory_curve(profile, time_metric='us', memory_metric='B', patch_for_step_plot=True)

        start_ts = profile.profiler_start_ts
        self.assertEqual(1, start_ts)

        curves = result['rows']

        self.assertIn('CPU', curves)
        self.assertIn('GPU0', curves)

        self.assertEqual(2 * len(self.event_data_cpu) - 1, len(curves['CPU']))
        for i in range(len(curves['CPU'])):
            if i % 2 == 0:  # original values
                # adjusted timestamp
                self.assertEqual(self.event_data_cpu[i//2][0] - start_ts,  curves['CPU'][i][0])
                # total allocated
                self.assertEqual(self.event_data_cpu[i//2][-2], curves['CPU'][i][1])
                # total reserved
                self.assertEqual(self.event_data_cpu[i//2][-1], curves['CPU'][i][2])
            else:  # interpolated values
                self.assertEqual(self.event_data_cpu[i//2+1][0] - start_ts,  curves['CPU'][i][0])
                self.assertEqual(self.event_data_cpu[i//2][-2], curves['CPU'][i][1])
                self.assertEqual(self.event_data_cpu[i//2][-1], curves['CPU'][i][2])

        self.assertEqual(2 * len(self.event_data_gpu) - 1, len(curves['GPU0']))
        for i in range(len(self.event_data_gpu)):
            if i % 2 == 0:  # original values
                self.assertEqual(self.event_data_gpu[i//2][0] - start_ts,  curves['GPU0'][i][0])
                self.assertEqual(self.event_data_gpu[i//2][-2], curves['GPU0'][i][1])
                self.assertEqual(self.event_data_gpu[i//2][-1], curves['GPU0'][i][2])
            else:  # interpolated values
                self.assertEqual(self.event_data_gpu[i//2+1][0] - start_ts,  curves['GPU0'][i][0])
                self.assertEqual(self.event_data_gpu[i//2][-2], curves['GPU0'][i][1])
                self.assertEqual(self.event_data_gpu[i//2][-1], curves['GPU0'][i][2])


class TestModuleView(unittest.TestCase):

    def test_build_module_hierarchy(self):
        from torch_tb_profiler.profiler import trace
        from torch_tb_profiler.profiler.module_op import (
            _build_module_hierarchy, aggegate_module_view)

        json_content = """[
            {
                "ph": "X",
                "cat": "python_function",
                "name": "test_root",
                "pid": 1908,
                "tid": 1908,
                "ts": 1,
                "dur": 19367,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 1,
                    "Python thread": 0
                }
            },
            {
                "ph": "X",
                "cat": "python_function",
                "name": "nn.Module: MyModule",
                "pid": 1908,
                "tid": 1908,
                "ts": 2,
                "dur": 211,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 2,
                    "Python parent id": 1,
                    "Python module id": 0
                }
            },
            {
                "ph": "X",
                "cat": "python_function",
                "name": "nn.Module: Linear",
                "pid": 1908,
                "tid": 1908,
                "ts": 5,
                "dur": 62,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 3,
                    "Python parent id": 2,
                    "Python thread": 0,
                    "Python module id": 1
                }
            },
            {
                "ph": "X",
                "cat": "cpu_op",
                "name": "aten::addmm",
                "pid": 1908,
                "tid": 1908,
                "ts": 10,
                "dur": 31,
                "args": {
                    "External id": 12182,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Fwd thread id": 0,
                    "Sequence number": 4006,
                    "python_caller_id": 3
                }
            },
            {
                "ph": "X",
                "cat": "python_function",
                "name": "nn.Module: MyModule",
                "pid": 1908,
                "tid": 1908,
                "ts": 1000,
                "dur": 211,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 4,
                    "Python parent id": 1,
                    "Python module id": 0
                }
            },
            {
                "ph": "X",
                "cat": "python_function",
                "name": "nn.Module: Linear",
                "pid": 1908,
                "tid": 1908,
                "ts": 1001,
                "dur": 62,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 5,
                    "Python parent id": 4,
                    "Python thread": 0,
                    "Python module id": 1
                }
            },
            {
                "ph": "X",
                "cat": "cpu_op",
                "name": "aten::addmm",
                "pid": 1908,
                "tid": 1908,
                "ts": 1002,
                "dur": 32,
                "args": {
                    "External id": 12182,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Fwd thread id": 0,
                    "Sequence number": 4006,
                    "python_caller_id": 5
                }
            },
            {
                "ph": "X",
                "cat": "python_function",
                "name": "nn.Module: MyModule",
                "pid": 1908,
                "tid": 1908,
                "ts": 2000,
                "dur": 211,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 6,
                    "Python parent id": 1,
                    "Python module id": 0
                }
            },
            {
                "ph": "X",
                "cat": "python_function",
                "name": "nn.Module: Linear",
                "pid": 1908,
                "tid": 1908,
                "ts": 2001,
                "dur": 62,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 7,
                    "Python parent id": 6,
                    "Python thread": 0,
                    "Python module id": 1
                }
            },
            {
                "ph": "X",
                "cat": "cpu_op",
                "name": "aten::addmm",
                "pid": 1908,
                "tid": 1908,
                "ts": 2002,
                "dur": 33,
                "args": {
                    "External id": 12182,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Fwd thread id": 0,
                    "Sequence number": 4006,
                    "python_caller_id": 7
                }
            },
            {
                "ph": "X",
                "cat": "python_function",
                "name": "nn.Module: Conv2",
                "pid": 1908,
                "tid": 1908,
                "ts": 3000,
                "dur": 211,
                "args": {
                    "External id": 0,
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "Python id": 8,
                    "Python parent id": 1,
                    "Python module id": 100
                }
            }
        ]
        """
        data = parse_json_trace(json_content)
        stats = aggegate_module_view(data.tid2tree, data.events)
        stats.sort(key=lambda x: x.name)
        self.assertEqual(2, len(stats))
        self.assertEqual('Conv2', stats[0].name)
        self.assertEqual('MyModule', stats[1].name)
        self.assertEqual(1, len(stats[1].children))
        self.assertEqual('Linear', stats[1].children[0].name)

        content = json.loads(json_content)

        events = []
        for data in content:
            event = trace.create_event(data, False)
            events.append(event)

        roots = _build_module_hierarchy(events)
        roots.sort(key=lambda x: x.name)
        self.assertEqual(2, len(roots))
        self.assertEqual('nn.Module: Conv2', roots[0].name)
        self.assertEqual('nn.Module: MyModule', roots[1].name)
        self.assertEqual(1, len(roots[1].children))
        self.assertEqual('nn.Module: Linear', roots[1].children[0].name)


class TestDataPipe(unittest.TestCase):

    def test_datapipe(self):
        json_content = """[
            {
                "ph": "X", "cat": "cpu_op",
                "name": "enumerate(DataPipe)#ShufflerIterDataPipe", "pid": 7557, "tid": 7557,
                "ts": 100, "dur": 23,
                "args": {
                    "External id": 34,
                    "Trace name": "PyTorch Profiler", "Trace iteration": 0
                }
            }
        ]"""
        profile = parse_json_trace(json_content)
        profile.process()

        dataloader_ranges = profile.role_ranges[ProfileRole.DataLoader]
        datapipe_range = None
        for range in dataloader_ranges:
            if range[0] == 100 and range[1] == 123:
                datapipe_range = range
                break
        self.assertTrue(datapipe_range is not None)

        root = next(iter(profile.tid2tree.values()))
        ops, _ = root.get_operator_and_kernels()
        datapipe_op = None
        for op in ops:
            if op.name.startswith('enumerate(DataPipe)'):
                datapipe_op = op
                break

        self.assertTrue(datapipe_op is None)


if __name__ == '__main__':
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
import os

from azure.storage.blob import ContainerClient

from .. import utils
from .base import BaseFileSystem, RemotePath, StatData
from .utils import as_bytes, as_text, parse_blob_url

logger = utils.get_logger()


class AzureBlobSystem(RemotePath, BaseFileSystem):
    """Provides filesystem access to S3."""

    def __init__(self):
        if not ContainerClient:
            raise ImportError('azure-storage-blob must be installed for Azure Blob support.')
        self.connection_string = os.environ.get('AZURE_STORAGE_CONNECTION_STRING', None)

    def exists(self, dirname):
        """Returns whether the path is a directory or not."""
        basename, parts = self.split_blob_path(dirname)
        if basename is None or parts is None:
            return False
        if basename == '':
            # root container case
            return True
        else:
            return basename == parts[0]

    def read(self, filename, binary_mode=False, size=None, continue_from=None):
        """Reads contents of a file to a string."""
        logger.info('azure blob: starting reading file %s' % filename)
        account, container, path = self.container_and_path(filename)
        client = self.create_container_client(account, container)
        blob_client = client.get_blob_client(path)
        if not blob_client.exists():
            raise FileNotFoundError("file %s doesn't exist!" % path)

        downloader = blob_client.download_blob(offset=continue_from, length=size)
        if continue_from is not None:
            continuation_token = continue_from + downloader.size
        else:
            continuation_token = downloader.size

        data = downloader.readall()
        logger.info('azure blob: file %s download is done, size is %d' % (filename, len(data)))
        if binary_mode:
            return as_bytes(data), continuation_token
        else:
            return as_text(data), continuation_token

    def write(self, filename, file_content, binary_mode=False):
        """Writes string file contents to a file."""
        account, container, path = self.container_and_path(filename)
        client = self.create_container_client(account, container)

        if binary_mode:
            if not isinstance(file_content, bytes):
                raise TypeError('File content type must be bytes')
        else:
            file_content = as_bytes(file_content)
        client.upload_blob(path, file_content)

    def download_file(self, file_to_download, file_to_save):
        logger.info('azure blob: starting downloading file %s as %s' % (file_to_download, file_to_save))
        account, container, path = self.container_and_path(file_to_download)
        client = self.create_container_client(account, container)
        blob_client = client.get_blob_client(path)
        if not blob_client.exists():
            raise FileNotFoundError("file %s doesn't exist!" % path)

        downloader = blob_client.download_blob()
        with open(file_to_save, 'wb') as downloaded_file:
            data = downloader.readall()
            downloaded_file.write(data)
            logger.info('azure blob: file %s is downloaded as %s, size is %d' %
                        (file_to_download, file_to_save, len(data)))

    def glob(self, filename):
        """Returns a list of files that match the given pattern(s)."""
        # Only support prefix with * at the end and no ? in the string
        star_i = filename.find('*')
        quest_i = filename.find('?')
        if quest_i >= 0:
            raise NotImplementedError(
                '{} not supported by compat glob'.format(filename)
            )
        if star_i != len(filename) - 1:
            return []

        filename = filename[:-1]

        account, container, path = self.container_and_path(filename)
        client = self.create_container_client(account, container)
        blobs = client.list_blobs(name_starts_with=path)
        return [blob.name for blob in blobs]

    def isdir(self, dirname):
        """Returns whether the path is a directory or not."""
        basename, parts = self.split_blob_path(dirname)
        if basename is None or parts is None:
            return False
        if basename == '':
            # root container case
            return True
        else:
            return basename == parts[0] and len(parts) > 1

    def listdir(self, dirname):
        """Returns a list of entries contained within a directory."""
        account, container, path = self.container_and_path(dirname)
        client = self.create_container_client(account, container)
        blob_iter = client.list_blobs(name_starts_with=path)
        items = []
        for blob in blob_iter:
            item = self.relpath(blob.name, path)
            if items not in items:
                items.append(item)
        return items

    def makedirs(self, dirname):
        """No need create directory since the upload blob will automatically create"""
        pass

    def stat(self, filename):
        """Returns file statistics for a given path."""
        account, container, path = self.container_and_path(filename)
        client = self.create_container_client(account, container)
        blob_client = client.get_blob_client(path)
        props = blob_client.get_blob_properties()
        return StatData(props.size)

    def walk(self, top, topdown=True, onerror=None):
        account, container, path = self.container_and_path(top)
        client = self.create_container_client(account, container)
        blobs = client.list_blobs(name_starts_with=path)
        results = {}
        for blob in blobs:
            dirname, basename = self.split(blob.name)
            dirname = 'https://{}/{}/{}'.format(account, container, dirname)
            results.setdefault(dirname, []).append(basename)
        for key, value in results.items():
            yield key, None, value

    def split_blob_path(self, blob_path):
        """ Find the first blob start with blob_path, then get the relative path starting from dirname(blob_path).
        Finally, split the relative path.
        return (basename(blob_path), [relative splitted paths])
        If blob_path doesn't exist, return (None, None)
        For example,
            For blob https://trainingdaemon.blob.core.windows.net/tests/test1/test2/test.txt
            * If the blob_path is '', return ('', [test1, test2, test.txt])
            * If the blob_path is test1, return (test1, [test2, test.txt])
            * If the blob_path is test1/test2, return (test2, [test2, test.txt])
            * If the blob_path is test1/test2/test.txt, return (test.txt, [test.txt])
        """
        account, container, path = self.container_and_path(blob_path)
        client = self.create_container_client(account, container)
        blobs = client.list_blobs(name_starts_with=path, maxresults=1)

        for blob in blobs:
            dir_path, basename = self.split(path)
            if dir_path:
                rel_path = blob.name[len(dir_path):]
                parts = rel_path.lstrip('/').split('/')
            else:
                parts = blob.name.split('/')
            return (basename, parts)
        return (None, None)

    def container_and_path(self, url):
        """Split an Azure blob -prefixed URL into container and blob path."""
        root, parts = parse_blob_url(url)
        if len(parts) != 2:
            raise ValueError('Invalid azure blob url %s' % url)
        return root, parts[0], parts[1]

    def create_container_client(self, account, container):
        if self.connection_string:
            client = ContainerClient.from_connection_string(self.connection_string, container)
        else:
            client = ContainerClient.from_container_url('https://{}/{}'.format(account, container))
        return client



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
from collections import namedtuple
from typing import Dict, List

OpAgg = namedtuple('OpAgg', [
    'name',
    'calls',
    'host_duration',
    'device_duration',
    'self_host_duration',
    'self_device_duration'])


class OpStats:
    def __init__(self,
                 name,
                 duration,
                 device_duration,
                 total_duration,
                 aggs: List[OpAgg]):
        self.name = name
        self.duration = duration
        self.device_duration = device_duration
        self.total_duration = total_duration
        self.op_aggs = aggs

    def __str__(self) -> str:
        return f'{self.name}: {self.duration}/{self.device_duration}/{len(self.op_aggs)}'


class DiffStats:
    def __init__(self, left: OpStats, right: OpStats):
        self.left = left
        self.right = right
        self.children: List[DiffStats] = []

    def flatten_diff_tree(self) -> Dict[str, 'DiffStats']:
        result: Dict[str, DiffStats] = {}

        def traverse(node: DiffStats, path: str):
            result[path] = node
            for i, child in enumerate(node.children):
                traverse(child, f'{path}-{i}')

        traverse(self, '0')
        return result

    def to_dict(self):
        d = {
            'left': {
                'name': self.left.name,
                'duration': self.left.duration,
                'device_duration': self.left.device_duration,
                'total_duration': self.left.total_duration,
                'aggs': []
            },
            'right': {
                'name': self.right.name,
                'duration': self.right.duration,
                'device_duration': self.right.device_duration,
                'total_duration': self.right.total_duration,
                'aggs': []
            }
        }

        for agg in self.left.op_aggs:
            d['left']['aggs'].append(agg._asdict())

        for agg in self.right.op_aggs:
            d['right']['aggs'].append(agg._asdict())

        return d

    def get_diff_tree_summary(self):
        def traverse_node_recursive(node: DiffStats):
            d = node.to_dict()

            d['children'] = []
            for c in node.children:
                d['children'].append(traverse_node_recursive(c))

            return d

        return traverse_node_recursive(self)

    def get_diff_node_summary(self, path: str):
        def traverse_node(node: DiffStats, path: str):
            d = node.to_dict()
            d['path'] = path
            return d

        d = traverse_node(self, path)
        d['children'] = []
        for i, c in enumerate(self.children):
            d['children'].append(traverse_node(c, f'{path}-{i}'))

        return d



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# --------------------------------------------------------------------------
from typing import Iterable, List

from .. import consts, utils
from .range_utils import (get_ranges_sum, intersection_ranges_lists,
                          intersection_ranges_lists_with_value, merge_ranges,
                          merge_ranges_with_value)
from .trace import BaseEvent, EventTypes, KernelEvent

logger = utils.get_logger()


# For calculating GPU utilization, and approximated SM efficiency.
class GPUMetricsParser(object):
    def __init__(self):
        # All gpu ids that used by any kernel.
        self.gpu_ids = set()
        # For calculating GPU utilization.
        self.kernel_ranges_per_device = [[] for _ in range(consts.MAX_GPU_PER_NODE)]
        self.gpu_utilization = [None] * consts.MAX_GPU_PER_NODE
        self.gpu_util_timeline_unit_size = 0
        self.gpu_util_timeline_unit_name = ''
        self.gpu_util_buckets = [[] for _ in range(consts.MAX_GPU_PER_NODE)]
        # For calculating approximated SM efficiency.
        self.blocks_per_sm_per_device = [[] for _ in range(consts.MAX_GPU_PER_NODE)]
        self.avg_approximated_sm_efficiency_per_device = [None] * consts.MAX_GPU_PER_NODE
        self.approximated_sm_efficiency_ranges = [[] for _ in range(consts.MAX_GPU_PER_NODE)]
        self.gpu_sm_efficiency_json = None
        self.blocks_per_sm_count = [0] * consts.MAX_GPU_PER_NODE
        # For calculating averaged occupancy.
        self.occupancy_per_device = [[] for _ in range(consts.MAX_GPU_PER_NODE)]
        self.avg_occupancy_per_device = [None] * consts.MAX_GPU_PER_NODE
        self.occupancy_count = [0] * consts.MAX_GPU_PER_NODE

    def calculate_gpu_utilization(self, global_start_time, global_end_time, steps_start_time, steps_end_time):
        # Make bucket_size to 10-power's of us, and number of buckets to (10, 100].
        # 10-power's of us, in order to straight forward for user to understand.
        # If number of buckets are too many, the value of gpu utilization will be either 0 or 1.
        def get_bucket_info(range_micro_seconds):
            max_buckets = 100
            bucket_size = 1
            while range_micro_seconds / bucket_size > max_buckets:
                bucket_size *= 10
            buckets = int(range_micro_seconds / bucket_size)
            unit = bucket_size
            unit_str = 'us'
            if unit >= 1000:
                unit /= 1000
                unit_str = 'ms'
                if unit >= 1000:
                    unit /= 1000
                    unit_str = 's'
            return int(bucket_size), int(buckets), int(unit), unit_str

        gpu_utilization_timeline = [[] for _ in range(consts.MAX_GPU_PER_NODE)]
        for gpu_id in self.gpu_ids:
            self.kernel_ranges_per_device[gpu_id] = merge_ranges(self.kernel_ranges_per_device[gpu_id])

            # Top-level number still consider steps, to be consistent with overview's breakdown.
            kernel_ranges_all_steps = intersection_ranges_lists(
                self.kernel_ranges_per_device[gpu_id], [(steps_start_time, steps_end_time)])
            ranges_sum = get_ranges_sum(kernel_ranges_all_steps)
            self.gpu_utilization[gpu_id] = ranges_sum / (steps_end_time - steps_start_time)

            # The timeline will use 'PyTorch Profiler (0)' as start,
            # in order to draw previous step's kernels' gpu utilization.
            bucket_size, buckets, self.gpu_util_timeline_unit_size, self.gpu_util_timeline_unit_name = \
                get_bucket_info(global_end_time - global_start_time)
            buckets_ranges = []
            for i in range(buckets):
                buckets_ranges.append((global_start_time + i * bucket_size,
                                       global_start_time + (i + 1) * bucket_size if i < buckets - 1
                                       else global_end_time))  # The last bucket may be longer.
            gpu_utilization_timeline[gpu_id] = [0] * buckets
            if len(self.kernel_ranges_per_device[gpu_id]) > 0:
                current_range_index = 0
                current_range = self.kernel_ranges_per_device[gpu_id][current_range_index]
                current_bucket_index = 0
                current_bucket = buckets_ranges[0]
                while (current_range_index < len(self.kernel_ranges_per_device[gpu_id])
                       and current_bucket_index < buckets):
                    if current_bucket[1] <= current_range[0]:
                        current_bucket_index += 1
                        current_bucket = buckets_ranges[current_bucket_index] if current_bucket_index < buckets \
                            else None
                    elif current_bucket[0] >= current_range[1]:
                        current_range_index += 1
                        if current_range_index < len(self.kernel_ranges_per_device[gpu_id]):
                            current_range = self.kernel_ranges_per_device[gpu_id][current_range_index]
                    else:
                        left_bound = max(current_range[0], current_bucket[0])
                        right_bound = min(current_range[1], current_bucket[1])
                        gpu_utilization_timeline[gpu_id][current_bucket_index] += (right_bound - left_bound)
                        if current_bucket[1] < current_range[1]:
                            current_bucket_index += 1
                            current_bucket = buckets_ranges[current_bucket_index] if current_bucket_index < buckets \
                                else None
                        else:
                            current_range_index += 1
                            if current_range_index < len(self.kernel_ranges_per_device[gpu_id]):
                                current_range = self.kernel_ranges_per_device[gpu_id][current_range_index]
                for i_bucket in range(buckets):
                    bucket_size = buckets_ranges[i_bucket][1] - buckets_ranges[i_bucket][0]
                    gpu_utilization_timeline[gpu_id][i_bucket] /= bucket_size
                    start_time = buckets_ranges[i_bucket][0]
                    self.gpu_util_buckets[gpu_id].append((start_time, gpu_utilization_timeline[gpu_id][i_bucket]))
                start_time = buckets_ranges[-1][1]
                self.gpu_util_buckets[gpu_id].append((start_time, 0))

        self.kernel_ranges_per_device = None  # Release memory.

    def calculate_approximated_sm_efficiency(self, steps_start_time, steps_end_time):
        def calculate_avg(approximated_sm_efficiency_ranges, total_dur):
            total_weighted_sm_efficiency = 0.0
            for r in approximated_sm_efficiency_ranges:
                dur = r[1] - r[0]
                total_weighted_sm_efficiency += r[2] * dur
            avg_approximated_sm_efficiency = total_weighted_sm_efficiency / total_dur
            return avg_approximated_sm_efficiency

        total_dur = steps_end_time - steps_start_time
        for gpu_id in self.gpu_ids:
            blocks_per_sm_ranges = self.blocks_per_sm_per_device[gpu_id]
            approximated_sm_efficiency_ranges = merge_ranges_with_value(blocks_per_sm_ranges)
            # To be consistent with GPU utilization, here it must also intersect with all steps,
            # in order to remove the kernels out of steps range.
            approximated_sm_efficiency_ranges_all_steps = intersection_ranges_lists_with_value(
                approximated_sm_efficiency_ranges, [(steps_start_time, steps_end_time)])
            if len(approximated_sm_efficiency_ranges_all_steps) > 0:
                avg_approximated_sm_efficiency = calculate_avg(approximated_sm_efficiency_ranges_all_steps, total_dur)
                self.avg_approximated_sm_efficiency_per_device[gpu_id] = avg_approximated_sm_efficiency

            # The timeline still uses all kernels including out of steps scope's.
            if len(approximated_sm_efficiency_ranges) > 0:
                self.approximated_sm_efficiency_ranges[gpu_id] = approximated_sm_efficiency_ranges

        self.blocks_per_sm_per_device = None  # Release memory.

    # Weighted average. Weighted by kernel's time duration.
    def calculate_occupancy(self, steps_start_time, steps_end_time):
        for gpu_id in self.gpu_ids:
            occupancys_on_a_device = self.occupancy_per_device[gpu_id]
            total_time = 0
            total_occupancy = 0.0
            for r in occupancys_on_a_device:
                min_time = max(r[0], steps_start_time)
                max_time = min(r[1], steps_end_time)
                if min_time < max_time:
                    dur = max_time - min_time
                    total_occupancy += r[2] * dur
                    total_time += dur
            if total_time > 0:
                self.avg_occupancy_per_device[gpu_id] = total_occupancy / total_time

    @classmethod
    def parse_events(cls,
                     events: Iterable[BaseEvent],
                     global_start_time: int,
                     global_end_time: int,
                     steps_start_time: int,
                     steps_end_time: int):
        parser = GPUMetricsParser()
        logger.debug('GPU Metrics, parse events')
        for event in events:
            if event.type == EventTypes.KERNEL:
                parser.parse_event(event)

        parser.calculate_gpu_utilization(global_start_time, global_end_time, steps_start_time, steps_end_time)
        parser.calculate_approximated_sm_efficiency(steps_start_time, steps_end_time)
        parser.calculate_occupancy(steps_start_time, steps_end_time)
        return parser

    def parse_event(self, event: KernelEvent):
        ts = event.ts
        dur = event.duration
        gpu_id = event.device_id
        if gpu_id != event.pid:
            logger.warning("pid '{}' is not equal to args.device '{}' on event with ts '{}'".format(
                event.pid, gpu_id, event.ts))
        if gpu_id is not None:
            if gpu_id not in self.gpu_ids:
                self.gpu_ids.add(gpu_id)
            self.kernel_ranges_per_device[gpu_id].append((ts, ts + dur))
            if event.blocks_per_sm is not None:
                if event.blocks_per_sm > 0.0:
                    self.blocks_per_sm_per_device[gpu_id].append((ts, ts + dur, event.blocks_per_sm))
                    self.blocks_per_sm_count[gpu_id] += 1
                else:
                    # Workaround for negative value input.
                    logger.warning('blocks per SM {} with ts {} is not positive!'.format(event.blocks_per_sm, ts))
            if event.occupancy is not None:
                if event.occupancy >= 0.0:
                    self.occupancy_per_device[gpu_id].append((ts, ts + dur, event.occupancy))
                    self.occupancy_count[gpu_id] += 1
                else:
                    # Workaround for negative value input.
                    logger.warning('est. achieved occupancy % {} with ts {} is negative!'.format(event.occupancy, ts))

    def get_gpu_metrics_columns(self):
        columns = []
        if self.has_blocks_per_sm:
            columns.append({'type': 'number', 'name': 'Mean Blocks Per SM',
                            'tooltip': consts.TOOLTIP_BLOCKS_PER_SM})
        if self.has_occupancy:
            columns.append({'type': 'number', 'name': 'Mean Est. Achieved Occupancy (%)',
                            'tooltip': consts.TOOLTIP_OCCUPANCY_COMMON + consts.TOOLTIP_OCCUPANCY_TABLE})
        return columns

    @property
    def has_blocks_per_sm(self):
        return sum(self.blocks_per_sm_count) > 0

    @property
    def has_occupancy(self):
        return sum(self.occupancy_count) > 0

    def get_gpu_metrics(self):
        def build_trace_counter_gpu_util(gpu_id, start_time, counter_value):
            util_json = ("{{\"ph\":\"C\", \"name\":\"GPU {} Utilization\", \"pid\":{}, \"ts\":{}, "
                         "\"args\":{{\"GPU Utilization\":{}}}}}").format(gpu_id, gpu_id, start_time, counter_value)
            return util_json

        def build_trace_counter_sm_efficiency(gpu_id, start_time, counter_value):
            util_json = ("{{\"ph\":\"C\", \"name\":\"GPU {} Est. SM Efficiency\", \"pid\":{}, \"ts\":{}, "
                         "\"args\":{{\"Est. SM Efficiency\":{}}}}}").format(gpu_id, gpu_id, start_time, counter_value)
            return util_json

        def add_trace_counter_gpu_util(gpu_id, start_time, counter_value, counter_json_list: List):
            json_str = build_trace_counter_gpu_util(gpu_id, start_time, counter_value)
            counter_json_list.append(json_str)

        def add_trace_counter_sm_efficiency(gpu_id, start_time, end_time, value, counter_json_list: List):
            efficiency_json_start = build_trace_counter_sm_efficiency(gpu_id, start_time, value)
            efficiency_json_finish = build_trace_counter_sm_efficiency(gpu_id, end_time, 0)
            counter_json_list.append(efficiency_json_start)
            counter_json_list.append(efficiency_json_finish)

        counter_json_list = []
        for gpu_id, buckets in enumerate(self.gpu_util_buckets):
            if len(buckets) > 0:
                # Adding 1 as baseline. To avoid misleading virtualization when the max value is less than 1.
                add_trace_counter_gpu_util(gpu_id, buckets[0][0], 1, counter_json_list)
                add_trace_counter_gpu_util(gpu_id, buckets[0][0], 0, counter_json_list)
            for b in buckets:
                add_trace_counter_gpu_util(gpu_id, b[0], b[1], counter_json_list)
        for gpu_id, ranges in enumerate(self.approximated_sm_efficiency_ranges):
            buckets = self.gpu_util_buckets[gpu_id]
            if len(ranges) > 0 and len(buckets) > 0:
                # Adding 1 as baseline. To avoid misleading virtualization when the max value is less than 1.
                add_trace_counter_sm_efficiency(gpu_id, buckets[0][0], buckets[0][0], 1, counter_json_list)
            for r in ranges:
                add_trace_counter_sm_efficiency(gpu_id, r[0], r[1], r[2], counter_json_list)

        return counter_json_list

    def get_gpu_metrics_data_tooltip(
            self,
            gpu_infos,
            tc_ratio):
        if not self.gpu_ids:
            return None, None

        has_sm_efficiency = False
        has_occupancy = False
        has_tc = False

        gpu_metrics_data = []
        gpu_info_columns = ['Name', 'Memory', 'Compute Capability']

        def process_gpu(gpu_id: int):
            nonlocal has_sm_efficiency, has_occupancy, has_tc
            gpu_metrics_data.append({'title': 'GPU {}:'.format(gpu_id), 'value': ''})
            gpu_info = gpu_infos.get(gpu_id, None)
            if gpu_info is not None:
                for key in gpu_info_columns:
                    if key in gpu_info:
                        gpu_metrics_data.append({'title': key, 'value': gpu_info[key]})
            else:
                # the legacy chrome tracing file would not have gpu info.
                pass
            gpu_metrics_data.append({'title': 'GPU Utilization', 'value': '{} %'.format(
                round(self.gpu_utilization[gpu_id] * 100, 2))})
            if self.avg_approximated_sm_efficiency_per_device[gpu_id] is not None:
                gpu_metrics_data.append({'title': 'Est. SM Efficiency', 'value': '{} %'.format(
                    round(self.avg_approximated_sm_efficiency_per_device[gpu_id] * 100, 2))})
                has_sm_efficiency = True
            if self.avg_occupancy_per_device[gpu_id] is not None:
                gpu_metrics_data.append({'title': 'Est. Achieved Occupancy', 'value': '{} %'.format(
                    round(self.avg_occupancy_per_device[gpu_id], 2))})
                has_occupancy = True
            if tc_ratio[gpu_id] is not None:
                gpu_metrics_data.append({'title': 'Kernel Time using Tensor Cores', 'value': '{} %'.format(
                    round(tc_ratio[gpu_id] * 100, 2))})
                has_tc = True

        gpu_ids = list(self.gpu_ids)
        process_gpu(gpu_ids[0])
        for idx in range(1, len(gpu_ids)):
            # Append separator line for beautiful to see.
            gpu_metrics_data.append({'title': '<hr/>', 'value': ''})
            process_gpu(gpu_ids[idx])

        tooltip_summary = 'The GPU usage metrics:\n'
        tooltip = '{}\n{}'.format(tooltip_summary,  consts.TOOLTIP_GPU_UTIL)
        if has_sm_efficiency:
            tooltip += '\n' + consts.TOOLTIP_SM_EFFICIENCY
        if has_occupancy:
            tooltip += '\n' + consts.TOOLTIP_OCCUPANCY_COMMON + consts.TOOLTIP_OCCUPANCY_OVERVIEW
        if has_tc:
            tooltip += '\n' + consts.TOOLTIP_TENSOR_CORES

        return gpu_metrics_data, tooltip



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# -------------------------------------------------------------------------
import tempfile

from .. import utils
from .. import multiprocessing as mp
from . import file
from .file import basename, is_local, download_file, read

logger = utils.get_logger()


class Cache:
    def __init__(self, cache_dir=None):
        self._lock = mp.Lock()
        self._manager = mp.Manager()
        self._cache_dict = self._manager.dict()
        self._cache_dir = cache_dir

    def __getstate__(self):
        """The multiprocessing module can start one of three ways: spawn, fork, or forkserver.
        The default mode is fork in Unix and spawn on Windows and macOS.
        Therefore, the __getstate__ and __setstate__ are used to pickle/unpickle the state in spawn mode.
        """
        data = self.__dict__.copy()
        # remove the _manager to bypass the following pickle error
        # TypeError: cannot pickle 'weakref' object
        if hasattr(self, '_manager'):
            del data['_manager']
        logger.debug('Cache.__getstate__: %s ' % data)
        return data, file._REGISTERED_FILESYSTEMS

    def __setstate__(self, state):
        """The default logging level in new process is warning. Only warning and error log can be written to
        streams.
        So, we need call use_absl_handler in the new process.
        """
        from absl import logging
        logging.use_absl_handler()
        logger.debug('Cache.__setstate__ %s ' % (state,))
        data, file._REGISTERED_FILESYSTEMS = state
        self.__dict__.update(data)

    def read(self, filename):
        local_file = self.get_remote_cache(filename)
        return read(local_file)

    @property
    def cache_dir(self):
        return self._cache_dir

    def get_remote_cache(self, filename):
        """Try to get the local file in the cache. download it to local if it cannot be found in cache."""
        local_file = self.get_file(filename)
        if local_file is None:
            if is_local(filename):
                return filename
            else:
                local_file = tempfile.NamedTemporaryFile(
                    'w+t', suffix='.%s' % basename(filename), dir=self._cache_dir, delete=False)
                local_file.close()
                download_file(filename, local_file.name)
                self.add_file(filename, local_file.name)
                return local_file.name

        return local_file

    def get_file(self, filename):
        return self._cache_dict.get(filename)

    def add_file(self, source_file, local_file):
        with self._lock:
            logger.debug('add local cache %s for file %s' % (local_file, source_file))
            self._cache_dict[source_file] = local_file

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._manager.__exit__(exc_type, exc_value, traceback)
